import copy
import os
import ssl
import time
from typing import Callable, Iterator, Literal

import backoff
from anthropic import (
    Anthropic,
    AnthropicBedrock,
    APIStatusError,
    InternalServerError,
    RateLimitError,
    Stream,
)
from anthropic.types import MessageStreamEvent
from boto3 import client as boto3_client
from httpx import Client, RemoteProtocolError
from loguru import logger
from openai import AzureOpenAI, OpenAI
from pydantic import BaseModel

from sweepai.config.server import (
    AWS_ACCESS_KEY,
    AWS_ANTHROPIC_AVAILABLE,
    AWS_ANTHROPIC_BASE_URL,
    AWS_REGION,
    AWS_SECRET_KEY,
    AZURE_API_KEY,
    AZURE_OPENAI_API_KEY,
    AZURE_OPENAI_API_VERSION,
    AZURE_OPENAI_ENDPOINT,
    CURL_CA_BUNDLE,
    DEEPSEEK_API_KEY,
    DEFAULT_GPT4_MODEL,
    MISTRAL_API_KEY,
    TOGETHER_API_KEY,
)
from sweepai.core.entities import AssistantMessage, Message, SystemMessage, UserMessage
from sweepai.core.llm.image_utils import create_message_with_images
from sweepai.core.llm.openai_proxy import OpenAIProxy
from sweepai.core.llm.viz_utils import save_messages_for_visualization
from sweepai.utils.cache import create_cache
from sweepai.utils.str_utils import (
    format_json_string,
    truncate_text_based_on_stop_sequence,
)
from sweepai.utils.streamable_functions import cached_streamable, streamable

cache = create_cache()
openai_proxy = OpenAIProxy()

OpenAIModel = Literal[
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-1106",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-16k-0613",
    "gpt-4-1106-preview",
    "gpt-4-0125-preview",
    "gpt-4-turbo-2024-04-09",
    "gpt-4-turbo",
    "gpt-4o",
    "gpt-4o-mini-2024-07-18",
    "gpt-4o-2024-08-06",
    "o1-preview",
    "o1-mini",
    "deepseek-coder",
    "meta.llama3-1-8b-instruct-v1:0",
    "meta.llama3-1-70b-instruct-v1:0",
    "meta.llama3-1-405b-instruct-v1:0",
    "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
    "mistral-large-latest",
]

AnthropicModel = Literal[
    "claude-3-haiku-20240307",
    "claude-3-sonnet-20240229",
    "claude-3-5-sonnet-20240620",
    "claude-3-5-sonnet-20241022",
    "claude-3-5-haiku-20241022",
    "claude-3-opus-20240229",
]

ChatModel = OpenAIModel | AnthropicModel

model_to_max_tokens: dict[ChatModel, int] = {
    "gpt-3.5-turbo": 4096,
    "gpt-3.5-turbo-1106": 16385,
    "gpt-3.5-turbo-16k": 16385,
    "gpt-4-1106-preview": 128000,
    "gpt-4-0125-preview": 128000,
    "gpt-4-turbo-2024-04-09": 128000,
    "gpt-4o": 128000,
    "gpt-4o-2024-08-06": 128000,
    "gpt-4o-mini-2024-07-18": 128000,
    "claude-v1": 9000,
    "claude-v1.3-100k": 100000,
    "claude-instant-v1.3-100k": 100000,
    "o1-preview": 128000,
    "o1-mini": 128000,
    "anthropic.claude-3-haiku-20240229-v1:0": 200000,
    "anthropic.claude-3-sonnet-20240229-v1:0": 200000,
    "claude-3-opus-20240229": 200000,
    "claude-3-sonnet-20240229": 200000,
    "claude-3-5-sonnet-20240620": 200000,
    "claude-3-5-sonnet-20241022": 200000,
    "claude-3-haiku-20240307": 200000,
    "claude-3-5-haiku-20241022": 200000,
    "gpt-3.5-turbo-16k-0613": 16000,
    "deepseek-coder": 128000,
    "meta.llama3-1-8b-instruct-v1:0": 128000,
    "meta.llama3-1-70b-instruct-v1:0": 128000,
    "meta.llama3-1-405b-instruct-v1:0": 128000,
    "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo": 128000,
    "mistral-large-latest": 128000,
}

default_temperature = 0.1

LATEST_CLAUDE_SONNET_MODEL = "claude-3-5-sonnet-20241022"
LATEST_CLAUDE_MODEL = "claude-3-5-haiku-20241022"

def is_openai_model(model: str) -> bool:
    return model.startswith("deepseek") or model.startswith("o1-") or model.startswith("gpt-4")


def is_rate_limit_or_overloaded_error(e):
    return isinstance(e, (RateLimitError, APIStatusError)) and e.status_code in (
        429,
        500,
        529,
    )


def wrap_anthropic_messages_with_cache(
    message_dicts: list[dict],
    indices: list[int] = [-2, -1, 1, 2],
):
    for index in indices:
        # check if index is in bounds here on both sides
        # the string type check is sufficient to prevent double wrapping
        index_valid = -len(message_dicts) <= index < len(message_dicts)
        if index_valid and isinstance(message_dicts[index]["content"], str):
            message_dicts[index]["content"] = [
                {
                    "type": "text",
                    "text": message_dicts[index]["content"],
                    "cache_control": {"type": "ephemeral"},
                }
            ]
    return message_dicts


# go through each message and see if we need to update it to include images or not
def add_images_to_messages(
    message_dicts: list[dict[str, str]],
    images: list[tuple[str, str, str]],
    use_openai: bool = False,
):
    if not images:
        return message_dicts
    try:
        new_message_dicts = []
        for message in message_dicts:
            new_message = create_message_with_images(message, images, use_openai=use_openai)
            new_message_dicts.append(new_message)
    except Exception as e:
        logger.error(f"Error in adding images to messages: {e}")
        return message_dicts
    return new_message_dicts


class Thread(BaseModel):
    messages: list[Message] = []
    model: ChatModel = DEFAULT_GPT4_MODEL
    file_change_paths: list[str] = []
    temperature: float = default_temperature

    class Config:
        arbitrary_types_allowed = True

    @property
    def messages_dicts(self):
        # Remove the key from the message object before sending to OpenAI
        cleaned_messages = [message.to_openai() for message in self.messages]
        return cleaned_messages

    @property
    def roles(self):
        return [message.role for message in self.messages]

    def __getitem__(self, item):
        return self.messages[item]

    def __iter__(self):
        return iter(self.messages)

    def __next__(self):
        return next(self.messages)

    def __len__(self):
        return len(self.messages)

    @property
    def last(self):
        return self.messages[-1]

    @property
    def anthropic_messages_dicts(self):
        """This method should never set state, only return the current state"""
        # Clean up multiple user messages in a row
        new_messages = []
        for message in self.messages_dicts:
            if message["role"] == "system":
                continue
            is_message_role_the_same = new_messages and new_messages[-1]["role"] == message["role"]
            if is_message_role_the_same:
                # this validates that the new message is not blank
                if message["content"].rstrip():
                    new_messages[-1]["content"] += "\n\n" + message["content"].rstrip()
            else:
                message["content"] = message["content"].rstrip()
                new_messages.append(copy.deepcopy(message))
        # We're seeing 400 errors from anthropic where the message is empty. Fixing this here but also adding a log
        for message in new_messages:
            if not message["content"]:
                message_order_log = [f'{message["role"]}\n{message["content"][:100]}' for message in new_messages]
                logger.error(f"Empty message found: {message}, message roles:\n{message_order_log}")
        # now parse out the empty messages, this may cause ANOTHER error because they aren't alternating, which we'll fix later
        new_messages = [message for message in new_messages if message["content"]]
        return new_messages

    @classmethod
    def from_system_message_string(cls, prompt_string: str, **kwargs):
        return cls(
            messages=[SystemMessage(content=prompt_string)],
            **kwargs,
        )

    @backoff.on_exception(
        wait_gen=backoff.expo,
        exception=(RateLimitError, APIStatusError),
        base=10,
        factor=2,
        max_time=40,
        max_tries=3,
        giveup=lambda e: not is_rate_limit_or_overloaded_error(e),
    )
    def chat(
        self,
        content: str,
        assistant_message_content: str = "",
        model: ChatModel = "claude-3-haiku-20240307",
        temperature: float | None = None,
        stop_sequences: list[str] = [],
        max_tokens: int = 4096,
        verbose: bool = True,
        images: list[tuple[str, str, str]] | None = None,
        stream: bool = False,
        __version__: int = 1,
    ) -> str | Iterator[str]:
        use_openai = is_openai_model(model)
        if use_openai:
            OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY")
            assert OPENAI_API_KEY or AZURE_API_KEY or AZURE_OPENAI_API_KEY
            max_tokens = 4096
            if self.model.startswith("gpt"):
                self.model = "gpt-4o"
            for message in self.messages:  # Not sure why this is needed
                if message.role == "system":
                    message.role = "user"
        else:
            ANTHROPIC_API_KEY = os.environ.get("ANTHROPIC_API_KEY")
            assert ANTHROPIC_API_KEY
            self.model = model
            if (AWS_ACCESS_KEY and AWS_SECRET_KEY and AWS_REGION) or model != LATEST_CLAUDE_MODEL:
                max_tokens = 4096
        if content:
            self.messages.append(Message(role="user", content=content))
        if assistant_message_content:
            self.messages.append(Message(role="assistant", content=assistant_message_content))
        temperature = temperature or self.temperature or default_temperature
        if verbose:
            # messages_string_to_log = '\n\n'.join([message.content for message in self.messages])
            if use_openai:
                logger.debug(f"Calling openai with model {model}\nInput:\n{content}")
            else:
                logger.debug(f"Calling anthropic with model {model}\nInput:\n{content}")
        system_message = "\n\n".join([message.content for message in self.messages if message.role == "system"])
        content = ""
        e = None
        use_aws = True

        def llm_stream():
            nonlocal model
            model = model or self.model
            streamed_text = "" if not assistant_message_content else assistant_message_content
            if use_openai:
                if model.startswith("deepseek"):
                    client = OpenAI(
                        api_key=DEEPSEEK_API_KEY,
                        base_url="https://api.deepseek.com",
                    )
                else:
                    if AZURE_OPENAI_API_KEY and AZURE_OPENAI_API_VERSION and AZURE_OPENAI_ENDPOINT:
                        http_client = None
                        if CURL_CA_BUNDLE:
                            http_client = Client(
                                verify=ssl.create_default_context(cafile=CURL_CA_BUNDLE),
                                proxies={"all://": None},
                            )
                        client = AzureOpenAI(
                            api_key=AZURE_OPENAI_API_KEY,
                            api_version=AZURE_OPENAI_API_VERSION,
                            azure_endpoint=AZURE_OPENAI_ENDPOINT,
                            http_client=http_client,
                        )
                        model = "gpt-4o-2024-05-13"
                    else:
                        client = OpenAI()
                if model.startswith("o1-"):
                    logger.info(f"Using o1 model {model}")
                    response = client.chat.completions.create(
                        model=model,
                        messages=self.messages_dicts,
                    )
                    streamed_text = (
                        response.choices[0].message.content
                        if not assistant_message_content
                        else assistant_message_content + response.choices[0].message.content
                    )
                    save_messages_for_visualization(
                        self.messages_dicts + [{"role": "assistant", "content": streamed_text}],
                        model_name=model,
                    )
                    yield streamed_text
                else:
                    try:
                        # add another one here for streaming because sometimes we end early
                        save_messages_for_visualization(self.messages_dicts, model_name=model)
                    except Exception as e:
                        logger.warning(f"Error saving messages for visualization: {e}")
                    response = client.chat.completions.create(
                        model=model,
                        messages=self.messages_dicts,
                        max_tokens=max_tokens,
                        temperature=temperature,
                        stream=True,
                    )
                    streamed_text = "" if not assistant_message_content else assistant_message_content
                    for chunk in response:
                        if not chunk.choices:
                            continue
                        new_content = chunk.choices[0].delta.content
                        if new_content:
                            streamed_text += new_content
                            yield new_content
                        if chunk.choices[0].finish_reason == "stop":
                            break
                        for stop_sequence in stop_sequences:
                            if stop_sequence in streamed_text:
                                response = truncate_text_based_on_stop_sequence(streamed_text, stop_sequences)
                                try:
                                    save_messages_for_visualization(
                                        self.messages_dicts + [{"role": "assistant", "content": response}],
                                        model_name=model,
                                    )
                                except Exception as e:
                                    logger.warning(f"Error saving messages for visualization: {e}")
                                return
                    if chunk.choices and chunk.choices[0] and chunk.choices[0].finish_reason:
                        logger.debug(f"finish_reason: {chunk.choices[0].finish_reason}")
            elif model.startswith("meta"):
                if TOGETHER_API_KEY:
                    from together import Together

                    client = Together(api_key=TOGETHER_API_KEY)
                    response = client.chat.completions.create(
                        model=model,
                        messages=self.messages_dicts,
                        max_tokens=min(max_tokens, 4097),
                        temperature=temperature,
                        stop=["<|eot_id|>"],
                        stream=True,
                    )
                    streamed_text = "" if not assistant_message_content else assistant_message_content
                    for chunk in response:
                        new_content = chunk.choices[0].delta.content
                        yield new_content

                        if any(stop_sequence in streamed_text for stop_sequence in stop_sequences):
                            return truncate_text_based_on_stop_sequence(streamed_text, stop_sequences)
                    return streamed_text
                else:
                    client = boto3_client(
                        "bedrock-runtime",
                        region_name=AWS_REGION,
                        aws_access_key_id=AWS_ACCESS_KEY,
                        aws_secret_access_key=AWS_SECRET_KEY,
                    )
                    message_dicts = self.anthropic_messages_dicts
                    message_dicts = [
                        {
                            "role": message["role"],
                            "content": [{"text": message["content"]}],
                        }
                        for message in message_dicts
                    ]

                    streaming_response = client.converse_stream(
                        modelId=model,
                        messages=message_dicts,
                        inferenceConfig={
                            "maxTokens": max_tokens,
                            "temperature": temperature,
                        },
                    )

                    streamed_text = "" if not assistant_message_content else assistant_message_content

                    for chunk in streaming_response["stream"]:
                        if "contentBlockDelta" in chunk:
                            text = chunk["contentBlockDelta"]["delta"]["text"]
                            streamed_text += text
                            yield text

                            if any(stop_sequence in streamed_text for stop_sequence in stop_sequences):
                                response = truncate_text_based_on_stop_sequence(streamed_text, stop_sequences)
                                save_messages_for_visualization(
                                    message_dicts + [{"role": "assistant", "content": response}],
                                    model_name=model,
                                )
                                return response
            elif model == "mistral-large-latest":
                from mistralai import Mistral

                client = Mistral(api_key=MISTRAL_API_KEY)

                chat_response = client.chat.stream(
                    model=model,
                    messages=self.messages_dicts,
                )

                streamed_text = "" if not assistant_message_content else assistant_message_content
                if chat_response is not None:
                    for event in chat_response:
                        streamed_text += event.data.choices[0].delta.content
                        yield event.data.choices[0].delta.content
                        if any(stop_sequence in streamed_text for stop_sequence in stop_sequences):
                            return truncate_text_based_on_stop_sequence(streamed_text, stop_sequences)
            else:
                if AWS_ANTHROPIC_AVAILABLE and use_aws:
                    if "anthropic" not in model:
                        if model == "claude-3-5-sonnet-20241022" or model == "claude-3-5-haiku-20241022":
                            model = "claude-3-5-sonnet-20240620" # sonnet v2 is not supported on bedrock
                        model = f"anthropic.{model}-v1:0"
                    if AWS_ANTHROPIC_BASE_URL:
                        request_headers = {"bedrock-key": ANTHROPIC_API_KEY}
                        client = AnthropicBedrock(
                            aws_access_key=AWS_ACCESS_KEY,
                            aws_secret_key=AWS_SECRET_KEY,
                            base_url=AWS_ANTHROPIC_BASE_URL,
                            http_client=Client(
                                verify=ssl.create_default_context(cafile=CURL_CA_BUNDLE),
                                headers=request_headers,
                            ),
                        )
                    else:
                        client = AnthropicBedrock(
                            aws_access_key=AWS_ACCESS_KEY,
                            aws_secret_key=AWS_SECRET_KEY,
                            aws_region=AWS_REGION,
                        )
                else:
                    client = Anthropic(
                        api_key=ANTHROPIC_API_KEY,
                    )
                start_time = time.time()
                message_dicts = self.anthropic_messages_dicts

                start_time = time.time()
                if verbose:
                    logger.info(f"In queue with model {model}...")
                streamed_text = "" if not assistant_message_content else assistant_message_content
                try:
                    if isinstance(client, AnthropicBedrock):
                        response: Stream[MessageStreamEvent] = client.messages.create(
                            model=model,
                            messages=message_dicts,
                            max_tokens=max_tokens,
                            temperature=temperature,
                            system=system_message,
                            stream=True,
                            # extra_headers={
                            #     "anthropic-beta": "max-tokens-3-5-sonnet-2024-07-15"
                            # } if not (ANTHROPIC_AVAILABLE and use_aws) else {}
                        )
                    else:
                        # PROMPT CACHING START
                        # add cache control to the messages (max 4 messages)
                        message_dicts = wrap_anthropic_messages_with_cache(
                            message_dicts,
                            indices=[
                                -2,
                                -1,
                                1,
                                2,
                            ],  # TODO: this strategy can be improved later by allowing the caller to pass in the indices
                        )
                        response = client.beta.prompt_caching.messages.create(
                            model=model,
                            messages=message_dicts,
                            max_tokens=max_tokens,
                            temperature=temperature,
                            system=system_message,
                            stream=True,
                            extra_headers={"anthropic-beta": "prompt-caching-2024-07-31"},
                        )
                        # PROMPT CACHING END
                    for event in response:
                        match event.type:
                            case "message_start":
                                cache_read_input_tokens = (
                                    event.message.usage.cache_read_input_tokens
                                    if hasattr(
                                        event.message.usage,
                                        "cache_read_input_tokens",
                                    )
                                    else 0
                                )
                                cache_creation_input_tokens = (
                                    event.message.usage.cache_creation_input_tokens
                                    if hasattr(
                                        event.message.usage,
                                        "cache_creation_input_tokens",
                                    )
                                    else 0
                                )
                                if verbose:
                                    logger.info(
                                        f"Starting stream with {event.message.usage.input_tokens} input tokens \nCACHE READ: {cache_read_input_tokens} and \nCACHE_WRITE: {cache_creation_input_tokens} with \n{len(message_dicts)} MESSAGES\nin {time.time() - start_time:.2f}s"
                                    )
                            case "content_block_delta":
                                streamed_text += event.delta.text
                                yield event.delta.text
                                if any(stop_sequence in streamed_text for stop_sequence in stop_sequences):
                                    if verbose:
                                        logger.info(f"Stop sequence hit after {time.time() - start_time:.2f}s")
                                    try:
                                        save_messages_for_visualization(
                                            self.anthropic_messages_dicts
                                            + [
                                                {
                                                    "role": "assistant",
                                                    "content": streamed_text,
                                                }
                                            ],
                                            model_name=model,
                                        )
                                        self.messages = self.messages + [AssistantMessage(content=streamed_text)]
                                    except Exception as e:
                                        logger.warning(f"Error saving messages for visualization: {e}")
                                    return truncate_text_based_on_stop_sequence(streamed_text, stop_sequences)
                            case "content_block_stop":
                                logger.info("")
                                break
                            case "content_block_start":
                                logger.info(event)
                            case _:
                                logger.info(event)
                except RemoteProtocolError as e:
                    if streamed_text and len(streamed_text) > 100:
                        logger.warning(f"RemoteProtocolError: {e}, probably Anthropic being flaky")
                    else:
                        raise
                except (APIStatusError, ValueError) as e:
                    # TODO: handle bedrock errors, also check for general AWS error
                    if isinstance(e, APIStatusError):
                        error_message = e.message
                    else:
                        error_message = str(e)
                    logger.warning(f"Error in Anthropic streaming response: {error_message}")
                    raise Exception(f"Anthropic API error: {error_message}") from e
                finally:
                    client.close()
                logger.info(f"Streamed {len(streamed_text)} characters in {time.time() - start_time:.2f}s")
            response_text = streamed_text
            try:
                save_messages_for_visualization(
                    self.anthropic_messages_dicts + [{"role": "assistant", "content": response_text}],
                    model_name=model,
                )
                self.messages = self.messages + [AssistantMessage(content=streamed_text)]
            except Exception as e:
                logger.warning(f"Error saving messages for visualization: {e}")
            return response_text

        if stream:
            return llm_stream()
        else:
            streamed_content = ""
            for new_token in llm_stream():
                streamed_content += new_token
            self.messages.append(AssistantMessage(content=streamed_content))
            if verbose:
                logger.debug(f"{model} response: {self.messages[-1].content}")
            try:
                save_messages_for_visualization(messages=self.anthropic_messages_dicts, model_name=model)
            except Exception as e:
                logger.exception(f"Failed to save messages for visualization due to {e}")
            return self.messages[-1].content


def format_string(s: str, params: dict) -> str:
    # Doing this to avoid format raising on missing inputs, raise on key error instead
    for key, value in params.items():
        placeholder = "{" + key + "}"
        if placeholder in s:
            s = s.replace(placeholder, str(value))
        else:
            raise KeyError(f"Key '{key}' not found in the string")
    return s


def call_llm(
    system_prompt: str,
    user_prompt: str,
    params: dict = {},
    *args,
    **kwargs,
):
    thread: Thread = Thread.from_system_message_string(
        prompt_string=system_prompt,
    )

    if params:
        user_prompt = format_string(user_prompt, params)
    return continuous_llm_calls(
        username=kwargs.get("username", "test_user"),
        content=user_prompt,
        thread=thread,
        *args,
        **kwargs,
    )


@cached_streamable(cache=cache)
def call_anthropic_with_word_buffer(thread: Thread, stop_sequences: list[str] = [], verbose: bool = True, **kwargs):
    # Watch out when using this because if it hits the cache, it won't enter the messages array
    response = ""
    yield ""
    for token in thread.chat(stop_sequences=stop_sequences, stream=True, **kwargs):
        if token:
            # yield the response after a certain point
            # I removed word buffer so we never close <new_code> and reopen the next <code_change>.
            response += token
            yield response
            if verbose:
                print(token, end="", flush=True)


@streamable
def call_anthropic_with_word_buffer_with_cache_handling(
    thread: Thread,
    stop_sequences: list[str] = [],
    content: str = "",
    **kwargs,
):
    thread_length = len(thread)
    response = ""
    for response in call_anthropic_with_word_buffer.stream(
        thread, stop_sequences=stop_sequences, content=content, **kwargs
    ):
        yield response
    if len(thread) == thread_length:
        # Simulates side effects when it hits the cache
        thread.messages.append(UserMessage(content=content))
        thread.messages.append(AssistantMessage(content=response))


def collapse_final_assistant_messages(thread: Thread) -> Thread:
    """
    Collapses the final contiguous block of assistant messages into a single message.
    The block ends when either:
    1. A non-assistant message is encountered
    2. The end of the thread is reached
    
    Example:
        [System, Assistant, Assistant, Assistant] -> [System, Assistant]
        [System, Assistant, Assistant, User] -> [System, Assistant, User]
        [System, Assistant, User, Assistant, Assistant] -> [System, Assistant, User, Assistant]
    
    Args:
        thread: Thread object containing a list of messages
        
    Returns:
        Thread with final contiguous assistant messages collapsed
        
    Raises:
        ValueError: If thread has fewer than 2 messages
    """
    messages = thread.messages
    if len(messages) < 2:
        raise ValueError("Thread must have at least 2 messages to collapse final assistant messages")

    # Find the last non-assistant message or start of thread
    last_idx = len(messages) - 1
    while last_idx >= 0 and messages[last_idx].role == "user":
        last_idx -= 1
        
    # No assistant messages to collapse
    if last_idx < 0 or messages[last_idx].role != "assistant":
        return thread
        
    # Find start of contiguous assistant block
    block_start = last_idx
    while block_start > 0 and messages[block_start - 1].role == "assistant":
        block_start -= 1
        
    # If no multi-message block found, return unchanged
    if block_start == last_idx:
        return thread
        
    # Combine content from messages in the contiguous block
    collapsed_content = "".join(
        message.content 
        for message in messages[block_start:last_idx + 1]
    )
    
    # Create new message list:
    # - Everything before the block
    # - Single collapsed assistant message
    # - Any trailing messages after the block
    collapsed_messages = (
        messages[:block_start] + 
        [Message(role="assistant", content=collapsed_content, key="assistant")] +
        messages[last_idx + 1:]
    )
    
    thread.messages = collapsed_messages
    return thread

def stream_backoff(
    stream_factory: Callable[[], Iterator],
    timeouts: list[float],
    exception_types: list[type[Exception] | str] = [Exception],
):
    error = None
    for timeout in timeouts:
        try:
            yield from stream_factory()
            break
        except Exception as e:
            logger.debug(f"Stream crashed due to {e}")
            for exception_type in exception_types:
                is_match = False
                if isinstance(exception_type, str):
                    # handle string case
                    is_match = exception_type in str(e)
                else:
                    # handle exception type case
                    is_match = isinstance(e, exception_type)
                if is_match:
                    error = e
                    break
            else:
                raise e
            logger.exception(f"Stream failed due to {error}")
            time.sleep(timeout)
    if error:
        raise error


def should_continue_calling_llm(
    latest_llm_generated_response: str,
    stop_sequences: list[str],
    num_calls: int,
    max_calls: int,
    max_token_output: int,
    model: str,
) -> bool:
    # stop if any stop sequence is present
    stop_sequence_present = any(token in latest_llm_generated_response for token in stop_sequences)
    if stop_sequence_present:
        return False
    # stop if we've called too many times
    if num_calls >= max_calls:
        return False
    # call it again if we've generated too many tokens without hitting a stop sequence
    # this previously used 80%, but i've seen the 3.5 factor be wrong by as much as 75%
    didnt_hit_max_tokens = len(latest_llm_generated_response) < (3.5 * max_token_output * 0.5)
    if didnt_hit_max_tokens:
        return False
    if model.startswith("o1-"):
        return False
    return True


@streamable
def continuous_llm_calls(
    username,
    thread: Thread,
    *args,
    stop_sequences: list[str] = ["</plan>"],
    MAX_CALLS=10,
    **kwargs,
):
    """
    Continuously call the LLM until the response is finished
    This will modify thread and will append the responses to the messages array as one singular large assistant message
    """
    response = ""
    for _ in range(3):
        for response in stream_backoff(
            lambda: call_anthropic_with_word_buffer_with_cache_handling.stream(
                thread,
                stop_sequences=stop_sequences,
                *args,
                **kwargs,
            ),
            timeouts=[0, 5, 5, 10, 30],
            exception_types=[InternalServerError, "Anthropic API Error", RemoteProtocolError, "The model produced invalid content"],
        ):
            yield response
        if response.strip():
            break
    else:
        logger.error(f"Failed to get first response {response}, last_message {thread.messages[-1]}")
    # initialize this to the current_response instead of ""
    # otherwise should_continue_calling_llm would immediately break
    next_response = response
    num_calls = 0
    max_token_output = kwargs.get("max_tokens", 4096)
    while should_continue_calling_llm(
        latest_llm_generated_response=next_response,
        stop_sequences=stop_sequences,
        num_calls=num_calls,
        max_calls=MAX_CALLS,
        max_token_output=max_token_output,
        model=kwargs.get("model", LATEST_CLAUDE_MODEL),
    ):
        content = ""
        response = response.rstrip()
        thread.messages[-1].content = response
        # ask for a second response
        try:
            if "content" in kwargs:
                kwargs.pop("content")
            next_response = ""

            for next_response in stream_backoff(
                lambda: call_anthropic_with_word_buffer_with_cache_handling.stream(
                    thread,
                    content=content,
                    stop_sequences=stop_sequences,
                    *args,
                    **kwargs,
                ),
                timeouts=[0, 5, 5, 10, 30],
                exception_types=[InternalServerError, "Anthropic API Error", RemoteProtocolError, "The model produced invalid content"],
            ):
                yield response + next_response
            thread = collapse_final_assistant_messages(thread)
            # We should break if this is empty because the LLM hasn't generated anything new
            if not next_response.strip():
                break
            # we can simply concatenate the responses
            response += next_response
        except Exception as e:
            logger.error(f"Failed to get second response due to {e}")
        num_calls += 1
    return response


# UNUSED RIGHT NOW - STRUCTURED OUTPUTS
def get_openai_structured_output(
    messages: list[Message],
    response_format: BaseModel,
    model: str = "gpt-4o-2024-08-06",
) -> BaseModel | None:
    """Returns structured output from OpenAI API, or None if the response is refused."""
    messages = [message.to_openai() for message in messages]
    messages_raw = "\n".join([message["content"] for message in messages])
    logger.debug(f"Calling OpenAI with structured output:\n{messages_raw}")

    def call_structured_output(
        model: str,
        messages: list[dict[str, str]],
        response_format: BaseModel,
    ) -> BaseModel:
        openai_client = OpenAI()
        completion = openai_client.beta.chat.completions.parse(
            model=model,
            messages=messages,
            response_format=response_format,
        )
        return completion

    completion = call_structured_output(model=model, messages=messages, response_format=response_format)
    logger.debug(f"OpenAI structured output:\n{format_json_string(completion.choices[0].message.parsed.json())}")
    if completion.choices[0].message.refusal:
        logger.error(f"OpenAI refused to generate structured output: {completion.choices[0].message.refusal}")
        return None
    return completion.choices[0].message.parsed



from dataclasses import dataclass, field


@dataclass
class CodeReviewIssue:
    file_name: str
    issue_description: str
    line_number: str

    def __hash__(self):
        return hash((self.file_name, self.issue_description, self.line_number))


@dataclass
class PRReviewComment:
    thread_id: str
    file_name: str
    line_number: int
    body: str
    is_resolved: bool
    is_outdated: bool
    author: str


@dataclass
class PRReviewCommentThread:
    thread_id: str
    file_name: str
    line_number: int
    is_resolved: bool
    is_outdated: bool
    comments: list[PRReviewComment]


@dataclass
class CodeReview:
    file_name: str
    diff_summary: str
    issues: list[CodeReviewIssue]
    potential_issues: list[CodeReviewIssue]


@dataclass
class CodeReviewByGroup:
    file_names: list[str]
    diff_summary: str
    issues: list[CodeReviewIssue]
    potential_issues: list[CodeReviewIssue]

    @property
    def group_name(self):
        return ",".join(self.file_names)

    def get_all_file_names(self):
        return ", ".join(self.file_names)


@dataclass
class Patch:
    file_name: str
    old_start: int
    old_count: int
    new_start: int
    new_count: int
    changes: str


@dataclass
class PRChange:
    file_name: str
    diff: str
    old_code: str
    new_code: str
    status: str
    patches: list[Patch]
    annotations: list[str] = field(default_factory=list)


@dataclass
class FunctionDef:
    file_name: str
    function_code: str
    start_line: int
    end_line: int


@dataclass
class GroupedFilesForReview:
    file_names: list[str]
    rendered_changes: str  # full rendered changes
    rendered_patches: str  # only has rendered patches
    rendered_source_code: str  # only has rendered patches

    @property
    def group_name(self):
        return ",".join(self.file_names)

    def get_group_name(self):
        return ",".join(self.file_names)

    def get_all_file_names(self):
        return ", ".join(self.file_names)

    def is_file_in_group(self, file_name):
        return file_name in self.file_names



import os
import re
from datetime import datetime, timezone
from typing import Iterator

from diskcache import Cache
from github import Github, UnknownObjectException
from github.PullRequestComment import PullRequestComment
from github.GithubObject import NotSet
from github.Issue import Issue
from github.Repository import Repository
from loguru import logger
from tqdm import tqdm

from sweepai.config.server import (
    CACHE_DIRECTORY,
    GITHUB_API_BASE_URL,
    GITHUB_BASE_URL,
    GITHUB_BOT_USERNAME,
)
from sweepai.core.github_links import CommentURL, IssueURL, parse_url
from sweepai.core.github_utils import REPO_CACHE_BASE_DIR, github_integration
from sweepai.utils.str_utils import ordered_dedup

METRICS_FILE = f"{CACHE_DIRECTORY}/metrics.csv"

metrics_cache = Cache(f"{CACHE_DIRECTORY}/metrics_cache")

metrics_cutoff_date = datetime(2024, 8, 1)

# TODO: this type of mapping exists elsewhere
FILE_EXTENSION_TO_LANGUAGE = {
    # Programming languages
    "py": "python",
    "ts": "typescript",
    "tsx": "typescript",
    "js": "javascript",
    "jsx": "javascript",
    "cpp": "cpp",
    "c": "c",
    "h": "c",
    "hpp": "cpp",
    "java": "java",
    "rb": "ruby",
    "php": "php",
    "go": "go",
    "rs": "rust",
    "swift": "swift",
    "kt": "kotlin",
    "scala": "scala",
    "cs": "csharp",
    "fs": "fsharp",
    "vb": "visualbasic",
    # Markup and data formats
    "md": "markdown",
    "txt": "plaintext",
    "yml": "yaml",
    "yaml": "yaml",
    "json": "json",
    "xml": "xml",
    "html": "html",
    "css": "css",
    "scss": "scss",
    "less": "less",
    # Configuration files
    "ini": "ini",
    "cfg": "config",
    "conf": "config",
    "toml": "toml",
    # Shell scripts
    "sh": "shell",
    "bash": "shell",
    "zsh": "shell",
    "bat": "batch",
    "ps1": "powershell",
    # Other
    "sql": "sql",
    "r": "r",
    "dockerfile": "dockerfile",
    "gitignore": "gitignore",
    "csv": "csv",
    "log": "log",
}


def make_aware(dt):
    if dt.tzinfo is None:
        return dt.replace(tzinfo=timezone.utc)
    return dt


# def get_bot_username():
#     return GITHUB_BOT_USERNAME.rstrip("[bot]") + "[bot]"


def extract_pr_url(content):
    pattern = r"https://.*?/pull/\d+"
    match = re.search(pattern, content, re.IGNORECASE)
    if match:
        return match.group(0)
    return None


def get_pr_links_from_issue(issue: Issue):
    """Works for issues that are labeled with sweep"""
    pr_links = []
    if issue.pull_request is None:
        comments = issue.get_comments()
        for comment in comments:
            if comment.user.login == GITHUB_BOT_USERNAME:
                pr_link = extract_pr_url(comment.body)
                if pr_link:
                    logger.info(f"Pull request URL found: {pr_link}")
                    pr_links.append(pr_link)
    return pr_links


def extract_issue_or_comment_url(content):
    # Try to find a link
    pattern = rf"https://{GITHUB_BASE_URL}.*?/(issues/\d+(#issuecomment-\d+)?|pull/\d+(#issuecomment-\d+|#discussion_r\d+|#r\d+))"
    match = re.search(pattern, content, re.IGNORECASE)
    if match:
        return match.group(0)
    return None


def get_issue_or_comment_link_from_pr(issue):
    if issue.pull_request is not None:
        issue_or_comment_link = extract_issue_or_comment_url(issue.body)
        if issue_or_comment_link:
            return issue_or_comment_link
    return None


def extract_issue_number_from_pr_body(content):
    pattern = r"Fixes(?:\s)*#(\d+)"
    match = re.search(pattern, content, re.IGNORECASE | re.MULTILINE)
    if match:
        return match.group(1)
    return None


def get_issue_link_from_pr_body(issue, repo):
    if issue.pull_request is not None:
        issue_number = extract_issue_number_from_pr_body(issue.body)
        if issue_number:
            issue = repo.get_issue(int(issue_number))
            return issue.html_url
    return None


def get_timestamp_of_last_label(issue: Issue):
    paginated_list = issue.get_timeline().reversed
    paginated_list.per_page = 100
    for event in paginated_list:
        if event.event == "labeled" and event.raw_data.get("label", {}).get("name") == "sweep":
            return make_aware(event.created_at)
    return datetime.fromtimestamp(0, tz=timezone.utc)  # same type, but 0 to indicate something is wrong


def get_original_issues_from_pr_link(
    g: Github, repo_name: str, ignore_cache: bool = False
) -> dict[str, (str, str, str, str)]:
    """Returns a dict where key is the pr_link and value is a tuple of (issue_link, issue_author, issue_title, issue.created_at)"""
    key = ("original_issues", repo_name, "v1")
    if key in metrics_cache and not ignore_cache:
        return metrics_cache.get(key)

    result = {}
    try:
        # Get the repository
        repo = g.get_repo(repo_name)
    except Exception as e:
        logger.error(f"Error accessing repository: {e}")
        return result

    try:
        key = ("issues_with_sweep_label", repo_name, "v0")
        if key not in metrics_cache or ignore_cache:
            issues_with_sweep_label: list[Issue] = ordered_dedup(
                list(
                    repo.get_issues(
                        labels=["sweep"],
                        state="all",
                        sort="created",
                        direction="desc",
                        since=metrics_cutoff_date,
                    )
                )
                + list(
                    repo.get_issues(
                        creator=f"{GITHUB_BOT_USERNAME}",
                        state="all",
                        sort="created",
                        direction="desc",
                        since=metrics_cutoff_date,
                    )
                )
            )
            metrics_cache.set(key, issues_with_sweep_label)
        else:
            issues_with_sweep_label: list[Issue] = metrics_cache.get(key)
        for issue in issues_with_sweep_label:
            logger.info(
                f"Issue #{issue.number}: {issue.title} Issue Author: {issue.user.login} Issue Updated At: {issue.created_at}"
            )
            pr_links = get_pr_links_from_issue(issue)
            logger.info(f"PR Links: {pr_links}")
            for pr_link in pr_links:
                if pr_link not in result:  # might want the latest here
                    result[pr_link] = (
                        issue.html_url,
                        issue.user.login,
                        issue.title,
                        issue.created_at,
                        get_timestamp_of_last_label(issue),
                    )
    except Exception as e:
        logger.error(f"Error accessing issues for repository {repo_name}: {e}")
        return result
    metrics_cache.set(key, result)
    return result


def retrieve_link_data(repo: Repository, issue_link: str):
    parsed_url = parse_url(issue_link)
    logger.debug(f"Issue link: {issue_link}")
    match parsed_url:
        case CommentURL(
            pull_request_number=issue_number,
            comment_id=comment_number,
            is_review_comment=is_review_comment,
        ):
            logger.debug(f"Issue comment")
            logger.debug(f"Issue number: {issue_number}")
            logger.debug(f"Issue comment number: {comment_number}")
            referenced_pr = repo.get_pull(int(issue_number))
            if is_review_comment:
                comment = referenced_pr.get_comment(int(comment_number))
            else:
                referenced_pr = referenced_pr.as_issue()
                comment = referenced_pr.get_comment(int(comment_number))
            issue_author = comment.user.login
            issue_title = comment.body
            issue_created_at = comment.created_at
            issue_labelled_at = issue_created_at
        case IssueURL(issue_number=issue_number):
            logger.debug(f"Issue")
            referenced_issue = repo.get_issue(issue_number)
            issue_author = referenced_issue.user.login
            issue_title = referenced_issue.title
            issue_created_at = referenced_issue.created_at
            issue_labelled_at = get_timestamp_of_last_label(referenced_issue)
        case _:
            raise ValueError(f"Invalid issue link: {issue_link}")
    return issue_author, issue_title, issue_created_at, issue_labelled_at

# Comment Extraction

def get_sweep_bot_comments(repo: Repository) -> Iterator[PullRequestComment]:
    repo._requester.per_page = 100
    comments = repo.get_issues_comments(
        sort="created",
        direction="desc",
        since=datetime(2024, 10, 15),
    )
    for comment in list(tqdm(comments)):
        if not hasattr(comment, "pull_request_url"): # only get PR comments
            continue
        body = comment.body
        if "Hey, @" in body and "I finished working on this. Here are the changes I made:" in body:
            yield comment

def extract_commit_sha_from_sweep_bot_comment(body: str) -> str | None:
    # E.g. 🔄 Revert 91fcee84c37da688a368cbb3a9b47acd9eb3e22e?
    pattern = r"Revert (?P<sha>\w{40})\?"
    match = re.search(pattern, body)
    if match:
        return match.group("sha")
    return None

def extract_username_from_sweep_bot_comment(body: str) -> str | None:
    # E.g. Hey, @username, I finished working on this. Here are the changes I made:
    pattern = r"Hey, @(?P<username>.+), I finished working on this. Here are the changes I made:"
    match = re.search(pattern, body)
    if match:
        return match.group("username")
    return None

def get_commit_metrics_repo(
    g: Github,
    repo_name: str,
    start_date: datetime | type[NotSet] | None = NotSet,
    ignore_cache: bool = False,
):
    all_commit_data_key = ("all_commit_data", repo_name, "v1")
    if all_commit_data_key in metrics_cache and not ignore_cache:
        return metrics_cache.get(all_commit_data_key)

    try:
        # Get the repository
        repo = g.get_repo(repo_name)
        repo._requester.per_page = 100 # sets pagination to 100
    except Exception as e:
        logger.error(f"Error accessing repository: {e}")
        return []

    # Retrieve PRs since the specified start date
    try:
        issues_created_by_sweep = list(
            repo.get_issues(
                creator=f"{GITHUB_BOT_USERNAME}",
                state="all",
                sort="created",
                direction="desc",
                since=start_date,
            )
        )
    except Exception as e:
        logger.error(f"Error accessing issues for repository {repo_name}: {e}")
        return []

    commit_data = []
    commit_unique_ids = set()
    for pr in issues_created_by_sweep:
        logger.info(f"PR: {pr.html_url}")

        # get the data for the pull request
        if pr.pull_request is None:
            continue
        pull_request_html_url = pr.pull_request.html_url
        pull_request_number = pull_request_html_url.split("/")[-1]
        if not pull_request_number.isdigit():
            continue
        try:
            pull_request = repo.get_pull(int(pull_request_number))
        except Exception as e:
            logger.error(f"Error accessing pull request {pull_request_number} for repository {repo_name}: {e}")
            continue

        # get the commits for the pull request
        try:
            commits = list(pull_request.get_commits())
        except Exception as e:
            logger.error(
                f"Error accessing commits for pull request {pull_request_number} for repository {repo_name}: {e}"
            )
            continue

        is_merged = pull_request.merged
        merged_by = pull_request.merged_by
        if is_merged and merged_by is not None and "sweep" in merged_by.login.lower():
            logger.info(f"Pull request {pull_request_html_url} was merged by Sweep")
            continue

        # get the issue link from the pull request
        issue_link = get_issue_or_comment_link_from_pr(pr)
        if issue_link is None:
            issue_link = get_issue_link_from_pr_body(pr, repo)
        logger.info(f"Issue Link: {issue_link}")

        if issue_link is None:
            issue_link = "N/A"
            issue_author = "N/A"
            issue_title = "N/A"
            issue_created_at = "N/A"
            issue_labelled_at = "N/A"
        else:
            issue_author, issue_title, issue_created_at, issue_labelled_at = retrieve_link_data(repo, issue_link)

        for commit in commits:
            logger.info(f"Commit: {commit.sha}")
            sha = commit.sha

            # Get the author's username
            author = commit.author.login if commit.author else commit.commit.author.name
            date = commit.commit.author.date

            try:
                # Access the files changed in the commit
                files = list(commit.files)
            except Exception as e:
                logger.error(f"Error accessing files for commit {sha}: {e}")
                continue

            for file in files:
                # Create a unique ID based on SHA and file path
                unique_id = f"{sha}_{file.filename}"
                if unique_id in commit_unique_ids:
                    continue
                commit_unique_ids.add(unique_id)

                additions = file.additions
                deletions = file.deletions

                # Detect the language by examining the file content
                file_extension = file.filename.split(".")[-1]
                language = FILE_EXTENSION_TO_LANGUAGE.get(file_extension, file_extension)

                new_commit_data = {
                    "id": unique_id,
                    "sha": sha,
                    "file_name": file.filename,
                    "org_name": repo.owner.login,
                    "repo_name": repo.name,
                    "author": author,
                    "date": date,
                    "additions": additions,
                    "deletions": deletions,
                    "language": language,
                    "pr_number": pr.number,
                    "pr_state": pr.state,
                    "pr_link": pull_request_html_url,
                    "pr_is_merged": is_merged,
                    "pr_created_at": pr.created_at,
                    "issue_link": issue_link,
                    "issue_author": issue_author,
                    "issue_title": issue_title,
                    "issue_created_at": issue_created_at,
                    "issue_labelled_at": issue_labelled_at,
                }
                metrics_cache.set(("commit_data", unique_id), new_commit_data)
                commit_data.append(new_commit_data)
    
    for comment in get_sweep_bot_comments(repo):
        body = comment.body
        commit_sha = extract_commit_sha_from_sweep_bot_comment(body)
        if commit_sha is None:
            continue
        username = extract_username_from_sweep_bot_comment(body)
        if username is None:
            continue
        try:
            # Access the files changed in the commit
            commit = repo.get_commit(commit_sha)
            files = list(commit.files)
        except Exception as e:
            logger.error(f"Error accessing files for commit {sha}: {e}")
            continue

        pr_number = comment.pull_request_url.split("/")[-1]

        for file in files:
            file_extension = file.filename.split(".")[-1]
            commit_data.append({
                "id": f"{commit_sha}_comment",
                "sha": commit_sha,
                "file_name": file.filename,
                "org_name": repo.owner.login,
                "repo_name": repo.name,
                "author": username,
                "date": comment.updated_at,
                "additions": file.additions,
                "deletions": file.deletions,
                "language": FILE_EXTENSION_TO_LANGUAGE.get(file_extension, file_extension),
                "pr_number": pr_number,
                "pr_state": "N/A",
                "pr_link": "N/A",
                "pr_is_merged": True,
                "pr_created_at": commit.commit.author.date,
                "issue_link": comment.html_url, # TODO: Add original issue link
                "issue_author": username,
                "issue_title": "N/A", # TODO: extract from blockquote
                "issue_created_at": comment.created_at, # these are approximations
                "issue_labelled_at": comment.created_at,
            })

    metrics_cache.set(all_commit_data_key, commit_data)
    return commit_data


def list_git_dirs(REPO_CACHE_BASE_DIR):
    paths = []

    def has_git_in_base(dir_path):
        base_dir = os.path.join(dir_path, "base")
        if os.path.isdir(base_dir):
            for dirname in os.listdir(base_dir):
                dirpath = os.path.join(base_dir, dirname)
                git_dir = os.path.join(dirpath, ".git")
                if os.path.isdir(git_dir):
                    return True
        return False

    for child in os.listdir(REPO_CACHE_BASE_DIR):
        child_path = os.path.join(REPO_CACHE_BASE_DIR, child)
        if os.path.isdir(child_path):
            # Check if child_path/base/ANYDIRNAME/.git exists
            if has_git_in_base(child_path):
                relative_path = os.path.relpath(child_path, REPO_CACHE_BASE_DIR)
                paths.append(relative_path)
            # Check for grandchild directories
            for grandchild in os.listdir(child_path):
                grandchild_path = os.path.join(child_path, grandchild)
                if os.path.isdir(grandchild_path):
                    # Check if grandchild_path/base/ANYDIRNAME/.git exists
                    if has_git_in_base(grandchild_path):
                        relative_path = os.path.relpath(grandchild_path, REPO_CACHE_BASE_DIR)
                        paths.append(relative_path)
    return paths


def get_commit_metrics(access_token: str = None, ignore_cache: bool = False) -> list[dict]:
    if not access_token:
        logger.info("Access token not provided. Using unauthenticated GitHub API.")
    repo_names = list_git_dirs(REPO_CACHE_BASE_DIR)
    commit_data = []

    g = Github(access_token, base_url=GITHUB_API_BASE_URL) if access_token else None

    for repo_full_name in tqdm(repo_names, desc="Processing repositories"):
        if (current_g := g) is None:
            org_name, repo_name = repo_full_name.split("/")
            try:
                installation = github_integration.get_repo_installation(owner=org_name, repo=repo_name)
            except UnknownObjectException as e:
                logger.error(f"Error accessing installation for repository {repo_full_name}: {e}")
                continue
            current_g = installation.get_github_for_installation()

        if commit_data_repo := get_commit_metrics_repo(current_g, repo_full_name, metrics_cutoff_date, ignore_cache):
            commit_data.extend(commit_data_repo)
            logger.info(f"Commit data for repository {repo_full_name}: {len(commit_data_repo)} commits")
    convert_metrics_to_csv(commit_data)
    return commit_data


def convert_metrics_to_csv(commit_data: list[dict]):
    import pandas as pd

    df = pd.DataFrame(commit_data)
    if len(df) == 0:
        print("No commit data to convert to CSV")
        return

    # Check for required columns and add them with default values if missing
    required_columns = [
        "pr_link",
        "pr_number",
        "pr_state",
        "pr_is_merged",
        "pr_created_at",
        "issue_link",
        "issue_author",
        "issue_title",
        "issue_created_at",
        "issue_labelled_at",
        "file_name",
        "language",
        "additions",
        "deletions",
        "author",
        "date",
        "repo_name",
        "org_name",
    ]

    for column in required_columns:
        if column not in df.columns:
            print(f"Column '{column}' is missing. Adding it with default value 'N/A'")
            df[column] = "N/A"

    # Convert 'date' column to datetime if it's not already
    if "date" in df.columns and not pd.api.types.is_datetime64_any_dtype(df["date"]):
        df["date"] = pd.to_datetime(df["date"], errors="coerce")

    def process_group(group):
        num_files = group["file_name"].nunique()
        group["total_changes"] = group["additions"] + group["deletions"]

        # language with most number of additions + deletions
        language_totals = group.groupby("language")["total_changes"].sum()
        primary_language = language_totals.idxmax()

        # list of other languages
        all_languages = group["language"].unique().tolist()
        all_languages_str = ";".join(all_languages)

        authors = group["author"].dropna().unique()
        authors_str = ";".join(authors)

        # For other columns, take the first value
        first_row = group.iloc[0]

        if first_row["issue_labelled_at"] is None or first_row["issue_created_at"] == "N/A":
            issue_to_pr_time_in_seconds = "N/A"
        else:
            issue_to_pr_time_in_seconds = (first_row["pr_created_at"] - first_row["issue_labelled_at"]).total_seconds()

        return pd.Series(
            {
                "pr_number": first_row["pr_number"],
                "pr_state": first_row["pr_state"],
                "pr_link": first_row["pr_link"],
                "pr_is_merged": any(group["pr_is_merged"]),
                "pr_created_at": first_row["pr_created_at"],
                "issue_link": first_row["issue_link"],
                "issue_author": first_row["issue_author"],
                "issue_title": first_row["issue_title"],
                "issue_created_at": first_row["issue_created_at"],
                "issue_labelled_at": first_row["issue_labelled_at"],
                "issue_creation_seconds": issue_to_pr_time_in_seconds,
                "num_files": num_files,
                "total_lines_changed": group["total_changes"].sum(),
                "primary_language": primary_language,
                "all_languages": all_languages_str,
                "authors": authors_str,
                "repo_name": first_row["repo_name"],
                "org_name": first_row["org_name"],
            }
        )

    try:
        grouped = df.groupby("pr_link")[required_columns].apply(process_group).reset_index(drop=True)
        logger.info(f"Writing metrics to {METRICS_FILE}")
        grouped.to_csv(METRICS_FILE, index=False)
        logger.info(f"Metrics successfully written to {METRICS_FILE}")
    except Exception as e:
        logger.exception(f"Error in grouping or writing CSV: {str(e)}")
        logger.error("Columns in DataFrame: " + ", ".join(df.columns))


if __name__ == "__main__":
    commit_data = get_commit_metrics(None, ignore_cache=True)
    logger.info(f"Commit data: {commit_data}")



"""
This module is responsible for handling the check suite event, called from sweepai/api.py
"""

import base64
import fnmatch
import os
import re
import shlex
import subprocess
import time
import zipfile
from contextlib import contextmanager
from dataclasses import dataclass, field
from time import sleep
from typing import Iterator

import psutil
import requests
from github.CommitStatus import CommitStatus
from github.Repository import Repository
from loguru import logger

from sweepai.config.client import get_config_key_value
from sweepai.config.server import CIRCLE_CI_PAT, DOCKER_BUILDKIT, ENV
from sweepai.core.github_utils import ClonedRepo
from sweepai.dataclasses.check_status import CheckStatus
from sweepai.dataclasses.dockerfile_config import (
    DockerfileConfig,
    load_dockerfile_configs_from_path,
)
from sweepai.utils.streamable_functions import streamable
from sweepai.utils.timer import Timer

MAX_LINES = 500
LINES_TO_KEEP = 100
CIRCLECI_SLEEP_DURATION_SECONDS = 15

DOCKERFILE_CONFIG_LOCATION = os.path.join(os.getcwd(), "dockerfiles")

log_message = """GitHub actions yielded the following error.

{error_logs}

Fix the code changed by the PR, don't modify the existing tests."""


def get_dirs(zipfile: zipfile.ZipFile):
    return [file for file in zipfile.namelist() if file.endswith("/") and "/" in file]


def get_files_in_dir(zipfile: zipfile.ZipFile, dir: str):
    return [file for file in zipfile.namelist() if file.startswith(dir) and not file.endswith("/")]


def remove_ansi_tags(logs: str) -> str:
    return re.sub(r"\x1b\[[0-9;]*[a-zA-Z]", "", logs, flags=re.MULTILINE)


gha_prompt = """\
The below command yielded the following errors:
<command>
{command_line}
</command>
<errors>
{error_content}
</errors>
Here are the logs:
<logs>
{cleaned_logs_str}
</logs>"""


@contextmanager
def change_dir(destination):
    prev_dir = os.getcwd()
    os.chdir(destination)
    try:
        yield
    finally:
        os.chdir(prev_dir)


@contextmanager
def changes_applied(cloned_repo: ClonedRepo, file_changes: dict):
    prev_files = {}
    try:
        # Save the changes and apply them
        for file_path, file_contents in file_changes.items():
            try:
                with open(os.path.join(cloned_repo.repo_dir, file_path), "r+") as file:
                    prev_files[file_path] = file.read()
                    file.seek(0)
                    file.write(file_contents)
                    file.truncate()
            except FileNotFoundError:
                logger.warning(f"File {file_path} not found in repo {cloned_repo.repo_dir}")
        yield
    finally:
        # Restore the original files
        for file_path, file_contents in prev_files.items():
            with open(os.path.join(cloned_repo.repo_dir, file_path), "w") as file:
                file.write(file_contents)


@dataclass
class DockerContainer:
    container_id: str
    file_changes: dict = field(default_factory=dict)

    def exec(self, command: str):
        return subprocess.run(
            f"docker exec {self.container_id} sh -c {shlex.quote(command)}",
            shell=True,
            check=True,
            capture_output=True,
            text=True,
        )

    def exec_stream(self, command: str):
        return subprocess.Popen(
            f"docker exec {self.container_id} sh -c {shlex.quote(command)}",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

    def pull_file_changes(self):
        for file_path in self.file_changes:
            results = self.exec(f"cat {shlex.quote(file_path)}")
            self.file_changes[file_path] = results.stdout


@contextmanager
def running_docker_container(
    image_name: str,
    container_name: str,
    file_changes: dict,
    env_path: str = "",
) -> Iterator[DockerContainer]:
    # Helper function to start up container, apply changes, move to cwd, and remove container after it exits
    if env_path:
        run_command = f"docker run --env-file {env_path} --name {container_name} -d {image_name} sleep infinity"
    else:
        run_command = f"docker run --name {container_name} -d {image_name} sleep infinity"

    results = subprocess.run(run_command, shell=True, check=True, capture_output=True, text=True)
    container_id = results.stdout.strip()

    container = DockerContainer(container_id, file_changes)

    make_changes_command = ""
    for file_path, file_contents in file_changes.items():
        dir_path = os.path.dirname(file_path)
        if dir_path:
            # Create the directory if it doesn't exist
            make_changes_command += f"mkdir -p {shlex.quote(dir_path)} && "
        # Write the file contents
        encoded_contents = base64.b64encode(file_contents.encode()).decode()
        # Base64 encoding prevents \n from being interpreted as separate lines
        command = f"echo {shlex.quote(encoded_contents)} | base64 -d > {shlex.quote(file_path)}"
        make_changes_command += f"{command} && "
    make_changes_command = make_changes_command.removesuffix(" && ")

    container.exec(make_changes_command)

    try:
        yield container
    finally:
        container.pull_file_changes()
        subprocess.run(
            f"docker kill {container_id} && docker rm {container_id}",
            shell=True,
            check=True,
        )


@streamable
def get_failing_docker_logs(
    cloned_repo: ClonedRepo,
    file_changes: dict,
):
    """
    Input: ClonedRepo object
    Output:
    - logs, a string containing the logs of the failing docker container
    - image_name, a string containing the name of the docker image (for later cleanup)
    """
    # image name should be an input arg and cleaned up at the end
    dockerfile_config_path = os.path.join(DOCKERFILE_CONFIG_LOCATION, cloned_repo.repo_full_name, "config.json")
    if not os.path.exists(dockerfile_config_path):
        logger.warning(f"Docker not configured for {cloned_repo.repo_full_name}")
        return "", []
    dockerfile_configs = load_dockerfile_configs_from_path(dockerfile_config_path)

    if not dockerfile_configs:
        # this method should not be called if the dockerfile config is not present
        return "", ""

    # Get CPU usage
    cpu_percent = psutil.cpu_percent(interval=1)

    # Get memory usage
    memory = psutil.virtual_memory()
    memory_percent = memory.percent

    # Get disk usage
    disk = psutil.disk_usage("/")
    disk_percent = disk.percent

    # Log system resource usage
    logger.info(f"System resource usage: CPU: {cpu_percent}%, Memory: {memory_percent}%, Disk: {disk_percent}%")

    # Get the number of running Docker processes
    docker_processes = subprocess.check_output(["docker", "ps", "-q"], text=True).splitlines()
    num_docker_processes = len(docker_processes)

    # Check if resource usage is too high
    if ENV == "prod" and (cpu_percent > 50 or memory_percent > 50 or disk_percent > 80 or num_docker_processes > 3):
        logger.warning("System resources are critically low. Docker build may fail or perform poorly.")
        raise Exception("System resources are critically low. Try again later.")

    @streamable
    def run_dockerfile_config(
        dockerfile_config: DockerfileConfig,
        cloned_repo: ClonedRepo,
        file_changes: dict,
    ):
        image_name = dockerfile_config.image_name + "-" + os.path.basename(cloned_repo.repo_dir)[:8]
        container_name = dockerfile_config.container_name + "-" + str(hash(cloned_repo.repo_dir))[-8:]
        dockerfile_path = dockerfile_config.dockerfile_path
        env_path = os.path.join(
            os.path.join(
                os.getcwd(),
                DOCKERFILE_CONFIG_LOCATION,
                cloned_repo.repo_full_name,
                ".env",
            )
        )
        _env_exists = os.path.exists(env_path)
        dockerfile_path = os.path.join(
            os.getcwd(),
            DOCKERFILE_CONFIG_LOCATION,
            cloned_repo.repo_full_name,
            dockerfile_path,
        )
        logs = ""

        status: CheckStatus = {
            "message": "",
            "stdout": "",
            "succeeded": None,
            "status": "running",
            "llm_message": "",
            "container_name": container_name,
            "start_time": -1,
            "end_time": -1,
        }
        with Timer():
            try:
                # TODO: add !.git to dockerignore
                with change_dir(cloned_repo.repo_dir):
                    # Build the Docker image
                    if DOCKER_BUILDKIT == "0":
                        build_command = f"DOCKER_BUILDKIT={DOCKER_BUILDKIT} docker build -t {image_name} -f {dockerfile_path} --build-arg CODE_PATH=. ."
                    else:
                        build_command = f"DOCKER_BUILDKIT={DOCKER_BUILDKIT} docker build --progress=plain -t {image_name} -f {dockerfile_path} --build-arg CODE_PATH=. ."
                    status["message"] = "Building Docker image..."
                    yield status, file_changes
                    status["start_time"] = time.time()
                    with Timer():
                        # Use Popen to stream output
                        stdout = ""
                        with subprocess.Popen(
                            build_command,
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            text=True,
                        ) as proc:
                            for line in proc.stdout:
                                print(line, end="", flush=True)
                                stdout += line
                                status["stdout"] = stdout
                                yield status, file_changes
                    logger.info(f"Built Docker image {image_name}")
                    # Check if the build process failed
                    if proc.returncode != 0:
                        status["message"] = "Docker build failed"
                        status["status"] = "failure"
                        status["succeeded"] = False
                        yield status, file_changes
                        return (
                            f"Docker build failed with exit code {proc.returncode}",
                            image_name,
                        )

                    env_path = ""

                    with running_docker_container(image_name, container_name, file_changes, env_path) as container:
                        # Find and dedup all file paths relative to the working directory
                        if dockerfile_config.working_dir:
                            file_paths = sorted(
                                [
                                    key.replace(dockerfile_config.working_dir.rstrip("/"), ".")
                                    for key in file_changes.keys()
                                    if key.startswith(dockerfile_config.working_dir)
                                ]
                            )
                        else:
                            file_paths = sorted(file_changes.keys())
                        file_paths_string = " ".join(file_paths)
                        file_dirs = list(set([os.path.dirname(file_path) for file_path in file_paths]))
                        file_dirs_string = " ".join(file_dirs)
                        base_command = dockerfile_config.command.replace("$FILE_PATHS", file_paths_string).replace(
                            "$FILE_DIRS", file_dirs_string
                        )

                        if dockerfile_config.working_dir:
                            run_command = f"cd {shlex.quote(dockerfile_config.working_dir)} && {base_command}"
                        else:
                            run_command = base_command

                        status["message"] = "Running Docker image..."
                        yield status, file_changes

                        logger.info(f"Running Docker image {image_name}...")
                        with Timer():
                            # Use Popen to stream output
                            stdout = f"$ ({dockerfile_config.working_dir or '~'}) {base_command}\n\n"
                            status["stdout"] = stdout
                            yield status, file_changes
                            with container.exec_stream(run_command) as proc:
                                for line in proc.stdout:
                                    print(line, end="", flush=True)
                                    stdout += line
                                    status["stdout"] = stdout
                                    yield status, file_changes
                        status["end_time"] = time.time()
                        status["message"] = f"Checks passed" if proc.returncode == 0 else f"Checks failed"
                        status["status"] = "success" if proc.returncode == 0 else "failure"
                        status["succeeded"] = proc.returncode == 0
                        yield status, file_changes

                    if proc.returncode == 0:
                        file_changes = container.file_changes

                    return "", image_name
            except subprocess.CalledProcessError as e:
                # TODO: handle this case
                logs = e.stdout + e.stderr
                logger.error(f"Error running Docker image {image_name}: {logs}")
            gha_formatted_prompt = gha_prompt.format(
                command_line=dockerfile_config.command,
                error_content=logs,
                cleaned_logs_str="",
            )
            status["message"] = "Checks failed"
            status["succeeded"] = False
            status["stdout"] = logs
            status["llm_message"] = gha_formatted_prompt
            yield status, file_changes
            return gha_formatted_prompt, image_name

    # run dockerfile configs in parallel
    docker_logs = ""
    image_names = []
    statuses = []
    new_dockerfile_configs = []
    for dockerfile_config in dockerfile_configs:
        did_match = False
        for file_glob in dockerfile_config.match_files:
            if any(fnmatch.fnmatch(file_path, file_glob) for file_path in file_changes.keys()):
                did_match = True
                break
        if did_match:
            new_dockerfile_configs.append(dockerfile_config)
    dockerfile_configs = new_dockerfile_configs

    for dockerfile_config in dockerfile_configs:
        statuses.append(
            {
                "message": "Queued",
                "stdout": "",
                "succeeded": None,
                "status": "pending",
                "llm_message": "",
                "container_name": dockerfile_config.container_name + "-" + str(hash(cloned_repo.repo_dir))[-8:],
            }
        )
    for i, dockerfile_config in enumerate(dockerfile_configs):
        for status, file_changes in run_dockerfile_config.stream(dockerfile_config, cloned_repo, file_changes):
            statuses[i] = status
            yield statuses, file_changes
        if not status["succeeded"]:
            statuses[i + 1 :] = [
                {**status, "status": "cancelled", "message": "Check cancelled"} for status in statuses[i + 1 :]
            ]
            yield statuses, file_changes
            break  # stop at the first failing docker container
    # TODO: delete all images > 2 hours old
    return docker_logs, image_names


def delete_docker_images(docker_image_names: str):
    for image_name in docker_image_names:
        delete_command = f"docker rmi {image_name}"
        try:
            subprocess.run(delete_command, shell=True, check=True)
            logger.info(f"Deleted Docker image: {image_name}")
        except subprocess.CalledProcessError as e:
            logger.warning(f"Error deleting Docker image: {image_name}")
            logger.warning(f"Error message: {str(e)}")


def clean_gh_logs(logs_str: str):
    """
    TODO: this can be made without the XML and truncation
    """
    # Extraction process could be better
    log_list = logs_str.split("\n")
    truncated_logs = [log[log.find(" ") + 1 :] for log in log_list]
    logs_str = "\n".join(truncated_logs)
    # extract the group and delete everything between group and endgroup
    gha_pattern = r"##\[group\](.*?)##\[endgroup\](.*?)(##\[error\].*)"
    match = re.search(gha_pattern, logs_str, re.DOTALL)
    if not match:
        return "\n".join(logs_str.split("\n")[:MAX_LINES])
    command_line = match.group(1).strip()
    log_content = match.group(2).strip()
    error_line = match.group(3).strip()  # can be super long
    return clean_cicd_logs(
        command=command_line,
        error=error_line,
        logs=log_content,
    )


def clean_cicd_logs(command: str, error: str, logs: str):
    patterns = [
        # for docker
        "Already exists",
        "Pulling fs layer",
        "Waiting",
        "Download complete",
        "Verifying Checksum",
        "Pull complete",
        # For github
        "remote: Counting objects",
        "remote: Compressing objects:",
        "Receiving objects:",
        "Resolving deltas:",
        "[command]/usr/bin/git ",
        "Download action repository",
        # For python
        "Collecting",
        "Downloading",
        "Installing",
        "━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━",
        # For prettier
        "npm WARN EBADENGINE ",
        "npm WARN deprecated ",
        "prettier/prettier",
    ]
    cleaned_logs = [
        log.strip() for log in logs.split("\n") if not any(log.strip().startswith(pattern) for pattern in patterns)
    ]
    if len(cleaned_logs) > MAX_LINES:
        # return the first LINES_TO_KEEP and the last LINES_TO_KEEP
        cleaned_logs = cleaned_logs[:LINES_TO_KEEP] + ["..."] + cleaned_logs[-LINES_TO_KEEP:]
    cleaned_logs_str = "\n".join(cleaned_logs)
    error_content = ""
    if len(error) < 200000:
        error_content = f"""<errors>
{error}
</errors>"""
    cleaned_response = gha_prompt.format(
        command_line=command,
        error_content=error_content,
        cleaned_logs_str=cleaned_logs_str,
    )
    return cleaned_response


def get_circleci_job_details(job_number, project_slug, vcs_type="github"):
    # project_slug is the repo full name
    headers = {"Circle-Token": CIRCLE_CI_PAT}
    url = f"https://circleci.com/api/v1.1/project/{vcs_type}/{project_slug}/{job_number}"
    response = requests.get(url, headers=headers)
    return response.json()


# take a commit and return all failing logs as a list
def get_failing_circleci_log_from_url(circleci_run_url: str, repo_full_name: str):
    if not CIRCLE_CI_PAT:
        logger.warning("CIRCLE_CI_APIKEY not set")
        return []
    headers = {"Circle-Token": CIRCLE_CI_PAT}
    job_number = circleci_run_url.split("/")[-1]
    circleci_run_details = get_circleci_job_details(job_number, repo_full_name)
    steps = circleci_run_details["steps"]
    failing_steps = []
    failed_commands_and_logs = ""
    for step in steps:
        if step["actions"][0]["exit_code"] != 0:
            failing_steps.append(step)
    for step in failing_steps:
        actions = step["actions"]
        for action in actions:
            if action.get("status") != "failed":
                continue
            if "output_url" in action:
                log_url = action["output_url"]
                log_response = requests.get(log_url, headers=headers)
                log_response = log_response.json()
                # these might return in a different order; watch out
                log_message = log_response[0]["message"] if len(log_response) > 0 else ""
                error_message = log_response[1].get("message", "") if len(log_response) > 1 else ""
                log_message = remove_ansi_tags(log_message)
                error_message = remove_ansi_tags(error_message)
                command = action.get("bash_command", "No command found.")  # seems like this is the only command
                circle_ci_failing_logs = clean_cicd_logs(
                    command=command,
                    error=error_message,
                    logs=log_message,
                )
                if circle_ci_failing_logs:
                    failed_commands_and_logs += circle_ci_failing_logs + "\n"
    return failed_commands_and_logs


def get_failing_circleci_logs(
    repo: Repository,
    current_commit: str,
):
    # get the pygithub commit object
    all_logs = ""
    failing_statuses = []
    total_poll_attempts = 0
    # hacky workaround because circleci can have a setup that takes a long time, and we will report "success" because the setup has finished but the actual CI is still running
    logger.debug("Waiting for 60 seconds before polling for CircleCI status.")
    sleep(60)
    while True:
        commit = repo.get_commit(current_commit)
        status = commit.get_combined_status()
        # https://docs.github.com/en/rest/commits/statuses?apiVersion=2022-11-28#get-the-combined-status-for-a-specific-reference
        all_statuses: list[CommitStatus] = status.statuses
        # if all are success, break
        if all(status.state == "success" for status in all_statuses):
            failing_statuses = []
            logger.debug(f"Exiting polling for CircleCI as all statuses are success. Statuses were: {all_statuses}")
            break
        # if any of the statuses are failure, return those statuses
        failing_statuses = [status for status in all_statuses if status.state == "failure"]
        if failing_statuses:
            logger.debug(f"Exiting polling for CircleCI as some statuses are failing. Statuses were: {all_statuses}")
            break
        # if any of the statuses are pending, sleep and try again
        if any(status.state == "pending" for status in all_statuses):
            if total_poll_attempts * CIRCLECI_SLEEP_DURATION_SECONDS // 60 >= 60:
                logger.debug("Polling for CircleCI has taken too long, giving up.")
                break
            # wait between check attempts
            total_poll_attempts += 1
            logger.debug(f"Polling to see if CircleCI has finished... {total_poll_attempts}.")
            sleep(CIRCLECI_SLEEP_DURATION_SECONDS)
            continue
    # filter out statuses that are not allowed
    # this only executes if allowed_cicd_names is not None
    allowed_cicd_names = get_config_key_value(repo, "allowed_cicd_names")
    if allowed_cicd_names:
        failing_statuses = [
            status
            for status in failing_statuses
            if any(cicd_name in status.context.lower() for cicd_name in allowed_cicd_names)
        ]
    # done polling
    for status_detail in failing_statuses:
        # CircleCI run detected
        if "circleci" in status_detail.context.lower():
            failing_circle_ci_log = get_failing_circleci_log_from_url(
                circleci_run_url=status_detail.target_url, repo_full_name=repo.full_name
            )  # may be empty string
            if failing_circle_ci_log:
                all_logs += failing_circle_ci_log + "\n"
    return all_logs



import concurrent.futures
import copy
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor

from loguru import logger
from tqdm import tqdm

from sweepai.config.client import SweepConfig, get_blocked_dirs
from sweepai.core.entities import Snippet
from sweepai.core.github_utils import ClonedRepo
from sweepai.dataclasses.search_dataclasses import Document
from sweepai.dataclasses.separated_snippets import SeparatedSnippets
from sweepai.o11y.posthog_trace import posthog_trace
from sweepai.search.query.cohere_utils import llm_rerank_call
from sweepai.search.query.lexical_search import (
    compute_vector_search_scores,
    prepare_lexical_search_index,
    search_index,
)
from sweepai.search.query.multi_query import generate_multi_queries
from sweepai.utils.str_utils import pack_items_for_prompt
from sweepai.utils.streamable_functions import streamable
from sweepai.utils.timer import Timer

# the order here matters as the first match is used
code_snippet_separation_features = {
    "tools": {
        "prefix": [
            ".git/",
            ".github/",
            ".circleci/",
            ".travis/",
            ".jenkins/",
            "scripts/",
            "script/",
            "bin/",
        ],
        "suffix": [
            ".gitignore",
            ".dockerignore",
            "Dockerfile",
            "Makefile",
            "Rakefile",
            "Procfile",
            ".sh",
            ".bat",
            ".cmd",
        ],
        "substring": [],
    },
    "junk": {  # we will discard this and not show it to the LLM
        "prefix": [
            "node_modules/",
            ".venv/",
            "build/",
            "venv/",
            "patch/",
            "target/",
            "bin/",
            "obj/",
        ],
        "suffix": [
            ".cache",
            ".gradle",
            ".mvn",
            ".settings",
            ".lock",
            ".log",
            ".tmp",
            ".tmp/",
            ".tmp.lock",
            ".tmp.lock/",
        ],
        "substring": [
            ".egg-info",
            "package-lock.json",
            "yarn.lock",
            ".cache",
            ".gradle",
            ".mvn",
        ],
    },
    "dependencies": {
        "prefix": [".", "config/", ".github/", "vendor/"],
        "suffix": [
            ".cfg",
            ".ini",
            ".po",
            "package.json",
            ".toml",
            ".yaml",
            ".yml",
            "LICENSE",
            ".lock",
        ],
        "substring": [
            "requirements",
            "pyproject",
            "Gemfile",
            "Cargo",
            "pom.xml",
            "build.gradle",
        ],
    },
    "docs": {
        "prefix": ["doc", "example", "README", "CHANGELOG"],
        "suffix": [".txt", ".rst", ".md", ".html", ".1", ".adoc", ".rdoc"],
        "substring": ["docs/", "documentation/"],
    },
    "tests": {
        "prefix": ["tests/", "test/", "spec/"],
        "suffix": [
            ".spec.ts",
            ".spec.js",
            ".test.ts",
            ".test.js",
            "_test.py",
            "_test.ts",
            "_test.js",
            "_test.go",
            "Test.java",
            "Tests.java",
            "Spec.java",
            "Specs.java",
            "_spec.rb",
            "_specs.rb",
            ".feature",
            "cy.ts",
            "cy.js",
        ],
        "substring": [
            "tests/",
            "test/",
            "/test",
            "_test",
            "rspec",
            ".test",
            "testdata/",
            "e2e/",
        ],
    },
}
# otherwise it's tagged as source
# we can make a config category later for css, config.ts, config.js. so far config files aren't many.

type_to_percentile_floor = {  # lower gets more snippets
    "tools": 0.3,
    "dependencies": 0.3,
    "docs": 0.3,
    "tests": 0.3,
    "source": 0.15,  # very low floor for source code
}

# set this to 0.2 because 0.1 means irrelevant in previous ranking steps
type_to_score_floor = {
    "tools": 0.2,
    "dependencies": 0.2,
    "docs": 0.2,
    "tests": 0.2,
    "source": 0.2,
}

type_to_result_count = {
    "tools": 5,
    "dependencies": 5,
    "docs": 5,
    "tests": 15,
    "source": 30,
}

rerank_count = {
    "tools": 10,
    "dependencies": 10,
    "docs": 30,
    "tests": 30,
    "source": 50,  # have to decrease to 30 for Voyage AI
}


def clamp_values(d: dict, max_value: float = 0.05) -> dict:
    if not d:
        return {}
    max_val = max(d.values())
    if max_val <= 0:
        return {k: 0 for k in d}
    return {k: min(max_value, max(0, v * max_value / max_val)) for k, v in d.items()}


def separate_snippets_by_type(snippets: list[Snippet]) -> SeparatedSnippets:
    separated_snippets = SeparatedSnippets()
    for snippet in snippets:
        for type_name, separation in code_snippet_separation_features.items():
            if (
                any(snippet.file_path.startswith(prefix) for prefix in separation["prefix"])
                or any(snippet.file_path.endswith(suffix) for suffix in separation["suffix"])
                or any(substring in snippet.file_path for substring in separation["substring"])
            ):
                separated_snippets.add_snippet(snippet, type_name)
                snippet.type_name = type_name
                break
        else:
            separated_snippets.add_snippet(snippet, "source")
    return separated_snippets


def apply_adjustment_score(
    snippet_path: str,
    old_score: float,
):
    snippet_score = old_score
    file_path, *_ = snippet_path.rsplit(":", 1)
    file_path = file_path.lower()
    # Penalize numbers as they are usually examples of:
    # 1. Test files (e.g. test_utils_3*.py)
    # 2. Generated files (from builds or snapshot tests)
    # 3. Versioned files (e.g. v1.2.3)
    # 4. Migration files (e.g. 2022_01_01_*.sql)
    base_file_name = file_path.split("/")[-1]
    if not base_file_name:
        return 0
    num_numbers = sum(c.isdigit() for c in base_file_name)
    snippet_score *= (1 - 1 / len(base_file_name)) ** num_numbers
    return snippet_score


NUM_SNIPPETS_TO_RERANK = 100
VECTOR_SEARCH_WEIGHT = 2


@streamable
def multi_get_top_k_snippets(
    cloned_repo: ClonedRepo,
    queries: list[str],
    k: int = 15,
    do_not_use_file_cache: bool = False,  # added for review_pr
    use_repo_dir: bool = False,
    seed: str = "",  # for caches
    globs: list[str] = ["**/*"],
):
    """
    Handles multiple queries at once now. Makes the vector search faster.
    """
    yield "Fetching configs...", [], [], []
    sweep_config: SweepConfig = SweepConfig()
    blocked_dirs = get_blocked_dirs(cloned_repo.github_repo)
    sweep_config.exclude_dirs += blocked_dirs
    # repository_directory = cloned_repo.cached_dir
    # if use_repo_dir:
    # We keep going back and forth between the cached_dir and the repo_dir. I set it to be repo_dir
    # because cached_dir was pointing to main and we were on a branch
    repository_directory = cloned_repo.repo_dir
    with Timer() as timer:
        for message, snippets, lexical_index in prepare_lexical_search_index.stream(
            repository_directory,
            sweep_config,
            do_not_use_file_cache=do_not_use_file_cache,
            seed=seed,
            globs=globs,
        ):
            yield message, [], snippets, []
        if len(snippets) == 0:
            yield "No snippets found for index", [], [], []
            return
        logger.info(f"Lexical indexing took {timer.time_elapsed} seconds")
        for snippet in snippets:
            snippet.file_path = snippet.file_path[len(cloned_repo.repo_dir) + 1 :]
        yield "Searching lexical index...", [], snippets, []
        with Timer() as timer:
            content_to_lexical_score_list = [search_index(query, lexical_index) for query in queries]
        logger.info(f"Lexical search took {timer.time_elapsed} seconds")

    # with Timer() as timer:
    #     _ = FilesSummarizer(cloned_repo=cloned_repo)
    logger.info(f"File summaries took {timer.time_elapsed} seconds")

    yield "Finished lexical search and file summaries, performing vector search...", [], snippets, []

    with Timer() as timer:
        for message, files_to_scores_list in compute_vector_search_scores.stream(queries, snippets):
            yield message, [], snippets, []
    logger.info(f"Vector search took {timer.time_elapsed} seconds")
    for i, query in enumerate(queries):
        for snippet in tqdm(snippets):
            vector_score = files_to_scores_list[i].get(snippet.denotation, 0.04)
            snippet_score = 0.02
            if snippet.denotation in content_to_lexical_score_list[i]:
                # roughly fine tuned vector score weight based on average score
                # from search_eval.py on 50 test cases May 13th, 2024 on an internal benchmark
                snippet_score = (
                    content_to_lexical_score_list[i][snippet.denotation] + (vector_score * VECTOR_SEARCH_WEIGHT)
                ) / (VECTOR_SEARCH_WEIGHT + 1)
                content_to_lexical_score_list[i][snippet.denotation] = snippet_score
            else:
                content_to_lexical_score_list[i][snippet.denotation] = snippet_score * vector_score
            content_to_lexical_score_list[i][snippet.denotation] = apply_adjustment_score(
                snippet_path=snippet.denotation,
                old_score=content_to_lexical_score_list[i][snippet.denotation],
            )
    ranked_snippets_list = [
        sorted(
            snippets,
            key=lambda snippet: content_to_lexical_score[snippet.denotation],
            reverse=True,
        )[:k]
        for content_to_lexical_score in content_to_lexical_score_list
    ]
    yield "Finished hybrid search, currently performing reranking...", ranked_snippets_list, snippets, content_to_lexical_score_list


@streamable
def get_top_k_snippets(
    cloned_repo: ClonedRepo,
    query: str,
    k: int = 15,
    do_not_use_file_cache: bool = False,  # added for review_pr
    use_repo_dir: bool = False,
    seed: str = "",
    *args,
    **kwargs,
):
    # Kinda cursed, we have to rework this
    for (
        message,
        ranked_snippets_list,
        snippets,
        content_to_lexical_score_list,
    ) in multi_get_top_k_snippets.stream(
        cloned_repo,
        [query],
        k,
        do_not_use_file_cache=do_not_use_file_cache,
        use_repo_dir=use_repo_dir,
        seed=seed,
        *args,
        **kwargs,
    ):
        yield message, (ranked_snippets_list[0] if ranked_snippets_list else []), snippets, (
            content_to_lexical_score_list[0] if content_to_lexical_score_list else []
        )


def get_pointwise_reranked_snippet_scores(
    query: str,
    snippets: list[Snippet],
    snippet_scores: dict[str, float],
    directory_summaries: dict,
    code_block_type: str,
    NUM_SNIPPETS_TO_RERANK=100,
):
    rerank_scores = copy.deepcopy(snippet_scores)

    sorted_snippets = sorted(
        snippets,
        key=lambda snippet: rerank_scores[snippet.denotation],
        reverse=True,
    )

    MAX_RERANKED_SNIPPETS = 20  # ceiling on the number of snippets to rerank
    snippet_representations: list[Document] = []
    # used for snippets that are not the current set of top snippets.
    valid_file_paths = set()
    most_relevant_snippet_per_file_path = dict()
    for snippet in sorted_snippets:
        if snippet.denotation not in valid_file_paths:
            valid_file_paths.add(snippet.file_path)
            most_relevant_snippet_per_file_path[snippet.file_path] = snippet

    all_directory_summaries = dict()
    for snippet in sorted_snippets[:MAX_RERANKED_SNIPPETS]:
        representation = f"{snippet.get_snippet(add_lines=False, add_ellipsis=False)}"
        # subdirs are all possible parent directories of the snippet
        subdir_names = []
        for subdir_slice in range(1, len(snippet.file_path.split("/"))):
            subdir_names.append("/".join(snippet.file_path.split("/")[:subdir_slice]))
        for subdir in subdir_names:
            if subdir in directory_summaries:
                # check the cache directly
                directory_summary = directory_summaries[subdir]
                if directory_summary in all_directory_summaries:
                    continue
                all_directory_summaries[subdir] = directory_summary
        snippet_representations.append(Document(snippet_denotation=snippet.denotation, content=representation))
    # 12k takes ~3s and can rank 10 items
    # 3k still takes ~3s and can rank 6 items, the bottleneck is TTFT
    # rest of pipeline is .7s
    snippet_representations = pack_items_for_prompt(
        iterable=snippet_representations,
        string_function=lambda x: x.snippet_denotation + x.content,
        token_limit=12_000,
    )
    NUM_SNIPPETS_TO_RERANK = len(snippet_representations)
    logger.info(f"NUM_SNIPPETS_TO_RERANK: {NUM_SNIPPETS_TO_RERANK}")

    # this needs to happen before we update the scores with the (higher) Cohere scores
    snippet_denotations = set(snippet.denotation for snippet in sorted_snippets)
    len(snippet_denotations)

    MAX_VALUE = 0.05
    new_snippet_scores = clamp_values(
        {
            snippet_denotation: v
            for snippet_denotation, v in rerank_scores.items()
            if snippet_denotation in snippet_denotations
        },
        MAX_VALUE,
    )

    response = llm_rerank_call(
        query=query,
        documents=snippet_representations,
        all_directory_summaries=all_directory_summaries,
        code_block_type=code_block_type,
        most_relevant_snippet_per_file_path=most_relevant_snippet_per_file_path,
    )
    for result_document in response.results:
        new_snippet_scores[result_document.snippet_denotation] = apply_adjustment_score(
            snippet_path=result_document.snippet_denotation,
            old_score=result_document.relevance_score,
        )

    # override score with Cohere score
    for snippet in sorted_snippets[:NUM_SNIPPETS_TO_RERANK]:
        if snippet.denotation in new_snippet_scores:
            snippet.score = new_snippet_scores[snippet.denotation]
    return new_snippet_scores


def process_snippets(type_name, *args, **kwargs):
    # not ideal but written this way to support concurrent.futures and for loop
    snippets_subset = args[1] if len(args) >= 2 else kwargs.get("snippets", [])
    if not snippets_subset:
        return type_name, {}
    return type_name, get_pointwise_reranked_snippet_scores(*args, **kwargs)


@streamable
@posthog_trace
def multi_prep_snippets(
    cloned_repo: ClonedRepo,
    queries: list[str],
    k: int = 15,
    skip_reranking: bool = False,  # This is only for pointwise reranking
    skip_pointwise_reranking: bool = False,
    skip_analyze_agent: bool = True,
    NUM_SNIPPETS_TO_RERANK=100,
    globs: list[str] = ["**/*"],
    directory_summaries: dict = {},
    *args,
    **kwargs,
):
    """
    Assume 0th index is the main query.
    """
    if len(queries) > 1:
        logger.info("Using multi query...")
        for (
            message,
            ranked_snippets_list,
            snippets,
            content_to_lexical_score_list,
        ) in multi_get_top_k_snippets.stream(
            cloned_repo,
            queries,
            k * 3,
            globs=globs,  # k * 3 to have enough snippets to rerank
        ):
            yield message, []
        # Use RRF to rerank snippets
        rank_fusion_offset = 0
        content_to_lexical_score = defaultdict(float)
        for i, ordered_snippets in enumerate(ranked_snippets_list):
            for j, snippet in enumerate(ordered_snippets):
                content_to_lexical_score[snippet.denotation] += content_to_lexical_score_list[i][snippet.denotation] * (
                    1 / 2 ** (rank_fusion_offset + j)
                )
        ranked_snippets = sorted(
            snippets,
            key=lambda snippet: content_to_lexical_score[snippet.denotation],
            reverse=True,
        )[:k]
    else:
        for (
            message,
            ranked_snippets,
            snippets,
            content_to_lexical_score,
        ) in get_top_k_snippets.stream(cloned_repo, queries[0], k, globs=globs):
            yield message, ranked_snippets
    separated_snippets = separate_snippets_by_type(snippets)
    yield f"Retrieved top {k} snippets, currently reranking:\n", ranked_snippets
    if not skip_pointwise_reranking and not skip_reranking:
        all_snippets = []
        for type_name, snippets_subset in separated_snippets:
            if len(snippets_subset) == 0:
                continue
            separated_snippets.override_list(
                type_name,
                sorted(
                    snippets_subset,
                    key=lambda snippet: content_to_lexical_score[snippet.denotation],
                    reverse=True,
                )[: rerank_count[type_name]],
            )
        new_content_to_lexical_score_by_type = {}
        with Timer() as timer:
            try:
                with ThreadPoolExecutor(max_workers=len(separated_snippets)) as executor:
                    future_to_type = {
                        executor.submit(
                            process_snippets,
                            type_name,
                            queries[0],
                            snippets_subset,
                            content_to_lexical_score,
                            directory_summaries,
                            type_name,
                            rerank_count[type_name],
                        ): type_name
                        for type_name, snippets_subset in separated_snippets
                    }
                    for future in concurrent.futures.as_completed(future_to_type):
                        type_name = future_to_type[future]
                        new_content_to_lexical_score_by_type[type_name] = future.result()[1]
            except RuntimeError as e:
                # Fallback to sequential processing
                logger.warning(e)
                for type_name, snippets_subset in separated_snippets:
                    new_content_to_lexical_score_by_type[type_name] = process_snippets(
                        type_name,
                        queries[0],
                        snippets_subset,
                        content_to_lexical_score,
                        directory_summaries,
                        type_name,
                        rerank_count[type_name],
                    )[1]
        logger.info(f"Reranked snippets took {timer.time_elapsed} seconds")
        for type_name, snippets_subset in separated_snippets:
            new_content_to_lexical_scores = new_content_to_lexical_score_by_type[type_name]
            for snippet in snippets_subset:
                snippet.score = new_content_to_lexical_scores[snippet.denotation]
            # set all keys of new_content_to_lexical_scores to content_to_lexical_score
            for key in new_content_to_lexical_scores:
                content_to_lexical_score[key] = new_content_to_lexical_scores[key]
            snippets_subset = sorted(
                snippets_subset,
                key=lambda snippet: new_content_to_lexical_scores[snippet.denotation],
                reverse=True,
            )
            separated_snippets.override_list(attribute_name=type_name, new_list=snippets_subset)
            logger.info(f"Reranked {type_name}")
            # cutoff snippets at percentile
            logger.info("Kept these snippets")
            if not snippets_subset:
                continue
            top_score = snippets_subset[0].score
            logger.debug(f"Top score for {type_name}: {top_score}")
            max_results = type_to_result_count[type_name]
            filtered_subset_snippets = []
            for idx, snippet in enumerate(snippets_subset[:max_results]):
                percentile = 0 if top_score == 0 else snippet.score / top_score
                if percentile < type_to_percentile_floor[type_name] or snippet.score <= type_to_score_floor[type_name]:
                    break
                logger.info(f"{idx}: {snippet.denotation} {snippet.score} {percentile}")
                snippet.type_name = type_name
                filtered_subset_snippets.append(snippet)
            logger.info(f"Length of filtered subset snippets for {type_name}: {len(filtered_subset_snippets)}")
            all_snippets.extend(filtered_subset_snippets)
        # if there are no snippets because all of them have been filtered out we will fall back to adding the highest rated ones
        # only do this for source files
        if not all_snippets:
            for type_name, snippets_subset in separated_snippets:
                # only use source files unless there are none in which case use all snippets
                if type_name != "source" and separated_snippets.source:
                    continue
                max_results = type_to_result_count[type_name]
                all_snippets.extend(snippets_subset[:max_results])

        all_snippets.sort(key=lambda snippet: snippet.score, reverse=True)
        ranked_snippets = all_snippets[:k]
        yield "Finished reranking, here are the relevant final search results:\n", ranked_snippets
    else:
        ranked_snippets = sorted(
            snippets,
            key=lambda snippet: content_to_lexical_score[snippet.denotation],
            reverse=True,
        )[:k]
        yield "Finished reranking, here are the relevant final search results:\n", ranked_snippets
    return ranked_snippets


@streamable
def prep_snippets(
    cloned_repo: ClonedRepo,
    query: str,
    k: int = 15,
    skip_reranking: bool = False,
    use_multi_query: bool = False,
    *args,
    **kwargs,
) -> list[Snippet]:
    if use_multi_query:
        queries = [query, *generate_multi_queries(query)]
        yield "Finished generating search queries, performing lexical search...\n", []
    else:
        queries = [query]
    for message, snippets in multi_prep_snippets.stream(cloned_repo, queries, k, skip_reranking, *args, **kwargs):
        yield message, snippets
    return snippets


if __name__ == "__main__":
    import os

    from sweepai.config.server import CACHE_DIRECTORY
    from sweepai.core.github_utils import MockClonedRepo
    from sweepai.utils.timer import Timer

    repo_full_name = os.environ.get("REPO_NAME")  # can test using this # REPO_NAME=sweepai/sweep-internal
    QUERY = os.environ.get("QUERY")  # QUERY="Write unit tests for ClonedRepo in github_utils.py"
    org_name, repo_name = repo_full_name.split("/")
    cloned_repo = MockClonedRepo(
        _repo_dir=f"{CACHE_DIRECTORY}/repos/{org_name}/{repo_name}/base/dev",
        repo_full_name=repo_full_name,
    )

    with Timer() as timer:
        for message, snippets in multi_prep_snippets.stream(cloned_repo, [QUERY], k=15):
            print(message, snippets)
    print("Time taken:", timer.time_elapsed)



import re

from loguru import logger

from sweepai.core.entities import Message
from sweepai.core.llm.chat import LATEST_CLAUDE_MODEL, Thread, call_llm
from sweepai.dataclasses.file_modification_state import FileModificationState
from sweepai.handlers.create_pr import INSTRUCTIONS_FOR_REVIEW
from sweepai.utils.diff import generate_diff
from sweepai.utils.str_utils import BOT_SUFFIX, extract_xml_tag, truncate_commit_message
from sweepai.utils.ticket_rendering_utils import get_branch_diff_text

commit_message_system_prompt = """Create a concise, informative commit message for GitHub from the code changes.

The output must:
1. Use an imperative tone, such as "Add", "Update", "Remove".
2. Your commit message MUST BE <50 characters long to adhere to GitHub's message length restrictions.
3. Group similar operations where applicable to avoid redundancy.
Your message must be directly related to the file changes stated. Avoid additional interpretation or detail not present in the input.

You will recieve a series of file diffs that need to be described in a short commit message.

Return your commit message in the following xml format:

<thinking>
1. Explain what happened in the file diffs, focusing on the lines added and removed. 
2. Try to identify the ideal commit message that reflects the changes made, whether the changes are substantial or minor. Remember to keep the message concise and clear, focusing on the file operations performed. Again, the commit message must be less than 50 characters long.
</thinking>

<commit_message>
short and concise description of the file diffs for a GitHub commit, not exceeding 50 characters
</commit_message>"""

commit_message_user_prompt = """Below are a series of file diffs to create a github commit message for:

{file_diffs}"""

pr_summary_system_prompt = """You are a helpful, excellent developer who is creating a pull request for a feature or bug fix. You need to write a pull request description that the changes in this pull request. You will always describe changes from higher to lower level, describing the purpose and value and then the details of the changes."""

pr_summary_prompt = """\
Write a pull request description that reflects all changes in this pull request. Here is the issue that this pull request is addressing:
<github_issue>
{issue}
</github_issue>

Here are the changes:
<diffs>
{diffs}
</diffs>

Format your response using the following XML tags:
<pr_title>
Title of the pull request.
</pr_title>
<pr_description>
# Purpose
Briefly describe the purpose of this pull request.
# Description
Description of the functional changes made in this pull request.
# Summary
Concise bulleted description of the pull request. Markdown format `variables`, `files`, and `directories` like this.
</pr_description>"""

GHA_SUMMARY_START = "<!-- GHA_SUMMARY_START -->"
GHA_SUMMARY_END = "<!-- GHA_SUMMARY_END -->"
MAX_COMMIT_MESSAGE_LENGTH = 64


class PRSummaryBot(Thread):
    # get commit message based on the patches
    # if previous patches are passed in then generate the incremental commit message
    def get_commit_message(
        self,
        modify_files_dict: dict[str, dict[str, str]],
        renames_dict: dict[str, str] = {},
        previous_modify_files_dict: dict[str, dict[str, str]] = {},
    ):
        self.messages = [
            Message(
                role="system",
                content=commit_message_system_prompt,
            )
        ]
        file_diffs = ""
        if not previous_modify_files_dict:
            for file_name, file_data in modify_files_dict.items():
                file_diff = generate_diff(file_data["original_contents"], file_data["contents"])
                file_diffs += f"<file_diffs file='{file_name}'>\n{file_diff}\n</file_diffs>"
        else:
            if modify_files_dict:
                for file_name, file_data in modify_files_dict.items():
                    # use incremental diff, compare against previous file data
                    if file_name in previous_modify_files_dict:
                        previous_file_data = previous_modify_files_dict[file_name]
                        file_diff = generate_diff(previous_file_data["contents"], file_data["contents"])
                    else:
                        # use diff compare against original file data
                        file_diff = generate_diff(file_data["original_contents"], file_data["contents"])
                    file_diffs += f"<file_diffs file='{file_name}'>\n{file_diff}\n</file_diffs>"
        for file_name, new_file_name in renames_dict.items():
            file_diff = f"File {file_name} was renamed to {new_file_name}"
            file_diffs += f"<file_diffs file='{file_name}'>\n{file_diff}\n</file_diffs>"

        if not file_diffs.strip():
            return "No changes were made"
        formatted_user_prompt = commit_message_user_prompt.format(file_diffs=file_diffs)
        commit_message_response = self.chat(
            content=formatted_user_prompt,
            model="claude-3-haiku-20240307",
        )
        commit_message = self.extract_commit_message(commit_message_response)
        if len(commit_message) > 50:
            shorter_commit_prompt = f'Your commit message is too long and was truncated to "{commit_message[:50]}".\nWrite a shorter commit message than "{commit_message}" that is still coherent.\n<commit_message>'
            shorter_commit_message_response = self.chat(
                content=shorter_commit_prompt,
                model="claude-3-haiku-20240307",
            )
            commit_message = self.extract_commit_message(shorter_commit_message_response)
        if not commit_message:
            logger.error("Failed to extract commit message from response.")
            commit_message = f"feat: Updated {len(modify_files_dict or [])} files"[:50]
        return commit_message

    def extract_commit_message(self, response: str):
        commit_message = ""
        commit_message_pattern = r"<commit_message>(?P<commit_message>.*?)</commit_message>"
        commit_message_match = re.search(commit_message_pattern, response, re.DOTALL)
        if commit_message_match:
            commit_message = commit_message_match.group("commit_message").strip()
        return commit_message

    # pylint: disable=E0213
    def get_pull_request_summary(
        problem_statement,
        issue_number,
        repo,
        overrided_branch_name,
        pull_request,
        pr_changes,
    ):
        # change the body here
        diff_text = get_branch_diff_text(
            repo=repo,
            branch=pull_request.branch_name,
            base_branch=overrided_branch_name,
        )
        # attempt to generate description 3 times
        for attempt in [0, 1, 2]:
            pr_desc_response = call_llm(
                system_prompt=pr_summary_system_prompt,
                user_prompt=pr_summary_prompt,
                params={
                    "issue": problem_statement,
                    "diffs": diff_text,
                },
            )
            pr_title_matches = re.search(r"<pr_title>\n(.*?)\n</pr_title>", pr_desc_response, re.DOTALL)
            pr_desc_matches = re.search(
                r"<pr_description>\n(.*?)\n</pr_description>",
                pr_desc_response,
                re.DOTALL,
            )
            if pr_desc_matches is None or pr_title_matches is None and attempt == 2:
                return pr_changes
            else:
                new_description = pr_desc_matches.group(1)
                new_title = pr_title_matches.group(1)
                pr_changes.title = f"Sweep: {new_title}"
                pr_changes.body = (
                    f"{new_description}\n\nFixes"
                    f" #{issue_number}.\n\n---\n{GHA_SUMMARY_START}{GHA_SUMMARY_END}\n\n{INSTRUCTIONS_FOR_REVIEW}{BOT_SUFFIX}"
                )
        return pr_changes


pr_summary_for_chat_system_prompt = """You are a helpful, excellent developer who is creating a pull request for a feature or bug fix. You need to write a pull request description that the changes in this pull request. You will always describe changes from higher to lower level, describing the purpose and value and then the details of the changes."""

pr_summary_for_chat_prompt = """\
Write a pull request description that reflects all changes in this pull request, for the repository {repo_name}.

Here is the conversation history that led to the changes:
<conversation>
{conversation}
</conversation>

<diff>
{diffs}
</diff>{pr_template}

Format your response using the following XML tags:

<pr_title>
Title of the pull request.
</pr_title>
<pr_description>
# Purpose
Briefly describe the purpose of this pull request.
# Description
Description of the functional changes made in this pull request.
# Summary
Concise bulleted description of the pull request. Markdown format `variables`, `files`, and `directories` like this.
</pr_description>"""


def get_pr_summary_for_chat(
    repo_name: str,
    modify_files_dict: FileModificationState,
    pr_template_string: str = "",
):
    diff_text = "\n\n".join(
        [
            f"{file_name}:\n" + generate_diff(file_contents.original_contents, file_contents.new_contents)
            for file_name, file_contents in modify_files_dict.files.items()
        ]
    )
    response = call_llm(
        system_prompt=pr_summary_for_chat_system_prompt,
        user_prompt=pr_summary_for_chat_prompt,
        params={
            "repo_name": repo_name,
            "diffs": diff_text,
            "pr_template": (
                f"\n\nHere is the repo's PR template:\n<pr_template>\n{pr_template_string}\n</pr_template>"
                if pr_template_string
                else ""
            ),
        },
        model=LATEST_CLAUDE_MODEL,
    )
    title = extract_xml_tag(response, "pr_title")
    assert title, "Title is required"
    description = extract_xml_tag(response, "pr_description") or ""
    return title, description


commit_message_chat_system_prompt = """You are a helpful, excellent developer who is creating a commit message for a feature or bug fix. You need to write a commit message for the changes in this diff."""

commit_message_chat_prompt = """\
You are tasked with writing a commit message based the diff of the changes made to relevant files. Your goal is to accurately summarize the changes and provide a concise, informative commit message.

Carefully analyze the diff of the changes:

<diff>
{diffs}
</diff>

Format your response using the following XML tags:

<commit_message>
Short and concise description of the file diffs for a GitHub commit, not exceeding {MAX_COMMIT_MESSAGE_LENGTH} characters
</commit_message>
"""


def get_commit_message_for_diff(modify_files_dict: FileModificationState):
    diff_text = "\n\n".join(
        [
            f"{file_name}:\n" + generate_diff(file_contents.original_contents, file_contents.new_contents)
            for file_name, file_contents in modify_files_dict.files.items()
        ]
    )
    response = call_llm(
        system_prompt=commit_message_chat_system_prompt,
        user_prompt=commit_message_chat_prompt,
        params={
            "diffs": diff_text,
            "MAX_COMMIT_MESSAGE_LENGTH": MAX_COMMIT_MESSAGE_LENGTH,
        },
    )
    commit_message = extract_xml_tag(response, "commit_message") or (
        f"Edited " + ", ".join(list(modify_files_dict.files.keys()))
    )
    commit_message = truncate_commit_message(commit_message)
    return commit_message



import os
import re
from dataclasses import dataclass, field
from typing import Generator, Optional

from loguru import logger

from sweepai.core.entities import Message
from sweepai.core.github_utils import ClonedRepo, MockClonedRepo
from sweepai.core.llm.chat import (
    LATEST_CLAUDE_MODEL,
    Thread,
    collapse_final_assistant_messages,
    continuous_llm_calls,
)
from sweepai.dataclasses.file_modification_state import FileModificationState
from sweepai.search.agent.code_tree import CodeTree
from sweepai.search.agent.snippet_utils import merge_snippet_ranges
from sweepai.utils.diff import generate_diff
from sweepai.utils.str_utils import (
    extract_lines,
    find_match_line_index,
    find_match_with_leading_spaces,
    reset_rightmost_string,
    should_truncate,
    strip_triple_quotes_preserve_whitespace,
    truncate,
)

MIN_LINES_FOR_BOUNDARY_TRUNCATION = 300
MAX_STREAM_CHANGES = 60
STREAM_VERSION = 2


def truncate_on_boundary(haystack: str, needle: str, language: str) -> str:
    start_index = find_match_line_index(haystack, needle)

    if start_index == -1:
        return ""

    end_index = start_index + len(needle.splitlines())

    try:
        code_tree = CodeTree.from_code(haystack, language=language)
    except Exception as e:
        logger.warning(f"Could not parse code tree for language {language}:\n{str(e)}")
        return ""
    boundaries = code_tree.get_boundaries()
    for boundary in boundaries:
        suggestion_is_long_enough = boundary - start_index > MIN_LINES_FOR_BOUNDARY_TRUNCATION
        if start_index <= boundary <= end_index and suggestion_is_long_enough:
            return extract_lines(haystack, start_index, boundary).strip("\n")
    return ""


def postprocessed_truncate_on_boundary(haystack: str, needle: str, language: str):
    boundary_truncated_code = truncate_on_boundary(haystack, needle, language)
    if boundary_truncated_code:
        boundary_truncated_code += "\n```\n</original_code>\n<new_code>\n```"
        return boundary_truncated_code
    return ""

def strip_partial_closing_tags(code: str) -> str:
    tag = "```\n</original_code>"
    while not code.endswith(tag):
        tag = tag[:-1]
    if len(tag) > len("```\n</or"):
        code = code.removesuffix(tag)
    return code

END_OF_ORIGINAL_CODE_MARKER = "```\n</original_code>\n<new_code>\n```"

def get_replacement_original_code(current_file_contents: str, partial_original_code: str) -> str:
    # If we should truncate, we return new right side. Otherwise, returns empty string.
    if should_truncate(current_file_contents, partial_original_code):
        logger.info("Truncating original code")
        truncated_code = truncate(current_file_contents, partial_original_code) # make this handle indentation changes
        if truncated_code:
            *_, remaining_code = current_file_contents.split(truncated_code, 1)
        else:
            # the first line of code is nowhere to be found in the file
            remaining_code = current_file_contents
            # breakpoint() # Good spot for breakpoint
        remaining_code = remaining_code.removeprefix("\n") # removes first newline
        if remaining_code:
            shortest_non_whitespace_substring = ""
            for line in remaining_code.splitlines():
                # Inject up to the first non-whitespace line
                shortest_non_whitespace_substring += line + "\n"
                if line.strip() and len(truncated_code) + len(shortest_non_whitespace_substring) > len(partial_original_code):
                    break
            replacement_original_code = ""
            shortest_non_whitespace_substring = shortest_non_whitespace_substring.rstrip("\n")
            if not shortest_non_whitespace_substring:
                replacement_original_code += END_OF_ORIGINAL_CODE_MARKER
            else:
                if shortest_non_whitespace_substring == remaining_code:
                    # This hit end of file so we should stop the original code
                    shortest_non_whitespace_substring += "\n" + END_OF_ORIGINAL_CODE_MARKER
                if truncated_code:
                    replacement_original_code = truncated_code + "\n" + shortest_non_whitespace_substring
                else:
                    replacement_original_code = shortest_non_whitespace_substring
        else:
            # file is complete
            replacement_original_code = truncated_code + "\n" + END_OF_ORIGINAL_CODE_MARKER
        return replacement_original_code
    return ""

def detect_infinite_loop_in_code_in_suggested_responses(all_suggested_responses: list[str]) -> bool:
    """
    Detects if the model is stuck in an infinite loop generating code suggestions. We will check if a certain state occurs >= 3 times.
    """
    code_suggestion_to_count = {}
    for suggested_response in all_suggested_responses:
        if suggested_response not in code_suggestion_to_count:
            code_suggestion_to_count[suggested_response] = 1
        else:
            code_suggestion_to_count[suggested_response] += 1
        if code_suggestion_to_count[suggested_response] >= 3:
            return True
    return False

@dataclass
class PartialCodeChange:
    """
    A partial code change that is generated by the model, only file_path is required
    """

    file_path: str
    raw_match: str
    justification: Optional[str] = None
    original_code: Optional[str] = None
    new_code: Optional[str] = None
    completed: bool = False
    generating_original_code: bool = False
    generating_new_code: bool = False

    def __eq__(self, other):
        if not isinstance(other, PartialCodeChange):
            return NotImplemented

        return (
            self.file_path == other.file_path
            and self.raw_match == other.raw_match
            and self.justification == other.justification
            and self.original_code == other.original_code
            and self.new_code == other.new_code
            and self.completed == other.completed
            and self.generating_original_code == other.generating_original_code
            and self.generating_new_code == other.generating_new_code
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(
            (
                self.file_path,
                self.raw_match,
                self.justification,
                self.original_code,
                self.new_code,
                self.completed,
                self.generating_original_code,
                self.generating_new_code,
            )
        )

    def __repr__(self):
        # print if completed
        # if completed, print "Completed" and the diff
        # otherwise print the original code and new code so far
        if self.completed and self.original_code and self.new_code:
            return "Completed" + generate_diff(old_code=self.original_code, new_code=self.new_code)
        return f"Not Completed:\nOriginal Code:\n{self.original_code}\nNew Code:\n{self.new_code}"


@dataclass
class CodeChangeStreamState:
    """
    Specialized state management for generating code changes
    """

    # if we can find a way to remove the following two fields, we should, we need to to call continuous_llm_calls after consuming the suggested response
    username: str
    model: str

    # we need all the below fields
    cloned_repo: ClonedRepo
    thread: Thread
    current_file_path: str = ""  # Path to the file that is being generated
    starting_response: str = ""  # Starting response from the model (to compare against the current_response)
    current_response: str = (
        ""  # Current response from the model (expected to be partially generated as it is being streamed)
    )
    suggested_response: str = (
        ""  # Suggested response to the model based on the current_response (the entire assistant message)
    )
    modify_files_dict: FileModificationState = field(default_factory=FileModificationState)
    code_change_partial_regex: str = (
        r"<code_change>\s*<file_path>\n*(?P<filePath>[\s\S]+?)\n*</file_path>\s*(<justification>\n*(?P<justification>[\s\S]+?)\n*(</justification>)?)?\s*(<original_code>\n*(?P<originalCode>[\s\S]*?\s*)(</original_code>)?\n*)?(<new_code>\n*(?P<newCode>[\s\S]+?)\n*(</new_code>)?)?\s*($|</code_change>|</original_code>|</new_code>)"
    )
    error_message: str = ""  # Error message to be displayed to the user
    partial_code_change: Optional[PartialCodeChange] = None
    applied_code_changes: set[PartialCodeChange] = field(default_factory=set)  # set of previously applied changes
    _all_suggested_responses: list[str] = field(default_factory=list)

    # also try and get rid of these fields too since the only reason we need them is for continuous_llm_calls
    stop_sequences: list[str] = field(default_factory=list)

    # Additional fields for detecting large negative diffs
    large_negative_diff_threshold: float = 0.2  # 20% of the file deleted
    last_generated_diff_was_large_deletion: bool = False  # Flag to track if the last diff was large
    last_large_deletion_diff_file_path: str = ""  # File path of the last large deletion diff
    last_validated_code_change: PartialCodeChange | None = None

    def get_latest_contents(self, file_path: str, safe: bool = False) -> str:
        """
        Get the latest contents of the file based on the modify_files_dict, this should be how you fetch the latest contents of a file
        """
        try:
            if file_path in self.modify_files_dict:
                return self.modify_files_dict.get_new_contents(file_path)
            return self.cloned_repo.get_file_contents(os.path.normpath(file_path))
        except FileNotFoundError as e:
            if safe:
                return ""
            raise e

    def get_partial_code_changes(self) -> list[PartialCodeChange]:
        """
        Get the partial code changes based on the current_response
        """
        partial_code_changes: list[PartialCodeChange] = []
        for match in re.finditer(self.code_change_partial_regex, self.current_response):
            raw_match = match.group(0)
            file_path = match.group("filePath")
            # this is the safe way to default these values, and I've only defaulted these three and not file_path because that's a fatal error
            justification = match.group("justification") or ""
            original_code = match.group("originalCode") or ""
            new_code = match.group("newCode") or ""
            generating_original_code = "<original_code>" in raw_match and "```\n</origin" not in raw_match # full match sometimes is insufficient
            generating_new_code = "<new_code>" in raw_match and "</new_code>" not in raw_match
            completed = "</code_change>" in raw_match
            # do some post processing on the fields
            if file_path:
                file_path = file_path.strip()
            if original_code and generating_original_code:
                # use strip_triple_quotes to remove the triple quotes from the original code
                original_code = strip_triple_quotes_preserve_whitespace(original_code)
            if new_code:
                new_code = strip_triple_quotes_preserve_whitespace(new_code)

            if completed:
                original_code = strip_triple_quotes_preserve_whitespace(original_code).strip("\n")
                new_code = strip_triple_quotes_preserve_whitespace(new_code).strip("\n")
            partial_code_changes.append(
                PartialCodeChange(
                    file_path=file_path,
                    raw_match=raw_match,
                    justification=justification,
                    original_code=original_code,
                    new_code=new_code,
                    completed=completed,
                    generating_original_code=generating_original_code,
                    generating_new_code=generating_new_code,
                )
            )
        return partial_code_changes

    def handle_early_update_state_exit(self):
        """
        Properly handle an early exit from update_state_with_new_response, this MUST BE CALLED before you return in the function
        """
        self.suggested_response = ""
        self.current_file_path = ""
        self.partial_code_change = None

    def detect_large_negative_diff(self, file_path: str, original_code: str, new_code: str) -> bool:
        """
        Detects if applying new_code to file_path would result in a large negative diff.

        Args:
            file_path (str): Path to the file being modified.
            new_code (str): The proposed new code to apply.

        Returns:
            bool: True if a large negative diff is detected, False otherwise.
        """
        try:
            current_file_contents = self.get_latest_contents(file_path)
            if original_code not in current_file_contents:
                # TODO: This is a fatal error, we should handle this better
                logger.error(f"Original code not found in {file_path}. Cannot detect large negative diff.")
                return False
            replaced_file_contents = current_file_contents.replace(original_code, new_code)
            # Compute the diff
            diff = generate_diff(current_file_contents, replaced_file_contents)
            total_lines = len(current_file_contents.splitlines())
            deleted_lines = 0
            for line in diff:
                if line.startswith("-") and not line.startswith("---"):
                    deleted_lines += 1
                elif line.startswith("+") and not line.startswith("+++"):
                    deleted_lines -= 1
            # Calculate the percentage of lines deleted
            if total_lines == 0:
                return False
            deletion_percentage = deleted_lines / total_lines
            if deletion_percentage > self.large_negative_diff_threshold or deleted_lines > 40:
                logger.warning(
                    f"Large negative diff detected for {file_path}: {deleted_lines} ({deletion_percentage*100:.2f}%) lines deleted."
                )
                return True
            return False
        except Exception as e:
            logger.error(f"Error detecting large negative diff for {file_path}: {e}")
            return False

    def generate_reset_new_code(self, file_path: str):
        """
        Resets the new_code by deleting all code until the last tag and starting over.

        Args:
            file_path (str): Path to the file being modified.
        """
        try:
            # Find the last occurrence of <new_code> tag
            last_new_code_tag = self.current_response.rfind("<new_code>")
            if last_new_code_tag == -1:
                # this is actually an error, but we should never hit this case
                logger.error(f"<new_code> tag not found in {file_path}. Cannot reset new_code.")
                return ""
            # Delete all code until the last <new_code> tag
            reset_content = self.current_response[:last_new_code_tag]
            # Append the opening <new_code> tag to start fresh
            reset_content += "<new_code>\n```"
            logger.info(f"Reset new_code for {file_path}.")
            return reset_content
        except Exception as e:
            logger.error(f"Error resetting new_code for {file_path}: {e}")
            return ""

    def validate_deletion_diffs_and_suggest_response(self, file_path: str):
        """
        Handles scenarios when a large negative diff is detected.

        Args:
            file_path (str): Path to the file being modified.
            new_code (str): The proposed new code to apply.
        """
        if self.detect_large_negative_diff(
            file_path=file_path,
            original_code=self.partial_code_change.original_code,
            new_code=self.partial_code_change.new_code,
        ):
            # Skip validation in two cases
            # Case 1: If the last generated diff was a large deletion and the current one is similar
            # Case 2: Checked the same file path twice (ensures we make progress)
            if self.last_generated_diff_was_large_deletion or self.last_large_deletion_diff_file_path == file_path:
                # Accept the large negative diff as it's similar to the previous one
                logger.info(f"Accepting large negative diff for {file_path} as the model has generated this twice.")
                self.last_generated_diff_was_large_deletion = False  # Reset flag
            else:
                # Reset the new_code and set the flag
                self.suggested_response = self.generate_reset_new_code(file_path)
                # Set these two fields so if we encounter the same one, we'll accept it
                self.last_generated_diff_was_large_deletion = True
                self.last_large_deletion_diff_file_path = file_path
        else:
            self.last_generated_diff_was_large_deletion = False  # Reset flag

    def check_if_suggested_code_is_at_end_of_file(self, file_contents: str, suggested_code: str) -> str:
        """
        Check if the suggested_code is at the end of the file. Returns the trailing string that follows the suggested_code if it is at the end of the file
        """
        rstripped_file_contents = file_contents.rstrip()
        # if the suggested code is at the end of the file, update the suggested code to include the rest of the file
        if rstripped_file_contents.endswith(suggested_code.rstrip()) and file_contents.count(suggested_code) == 1:
            match_index = file_contents.index(suggested_code)
            matched_content = file_contents[match_index + len(suggested_code) :]
            trailing_content = f"{matched_content}\n```\n</original_code>"
            return trailing_content
        return ""

    def try_generate_original_code_with_entities(self, partial_original_code: str) -> str:
        """
        Attempt to generate the original code based on the entities in the file
        """
        MIN_ENTITY_SUGGESTION_LENGTH = 30
        MAX_ENTITY_SUGGESTION_LENGTH = 100  # when suggesting entities, we want to limit the length of the suggestion
        # note: during this experiment I decreased from 300 to 100. then i decreased from 100 to 50. because it leads to too many problems when you suggest huge chunks of code
        # having it be so long causes the LLM to delete code because original code is so long.
        language = self.current_file_path.split(".")[-1]
        current_file_contents = self.get_latest_contents(self.current_file_path)
        rest_of_original_code_to_suggest = ""
        try:
            # check if an entities in the code file match against the partial code change
            tree = CodeTree.from_code(current_file_contents, language=language)
            # could be optimized by calling a specially created function to only get the entities with the code already filled in but this works for now
            entities_spans = tree.get_entities_spans()
            if entities_spans:
                import_line_ranges = tree.get_import_line_ranges()
                merged_import_ranges = merge_snippet_ranges(import_line_ranges, context_lines=5)
                current_file_contents_lines = current_file_contents.splitlines()

                for start_line, end_line in merged_import_ranges:
                    entities_spans[start_line] = end_line

                matched_start_line = find_match_line_index(
                    current_file_contents, partial_original_code
                )  # -1 if not found so this is safe
                matched_end_line = matched_start_line + len(partial_original_code.splitlines())

                for line_index in range(matched_start_line, matched_end_line):
                    if line_index in entities_spans:
                        predicted_end_line = entities_spans[line_index]
                        if (
                            predicted_end_line > matched_end_line
                            and MIN_ENTITY_SUGGESTION_LENGTH
                            < predicted_end_line - matched_end_line
                            < MAX_ENTITY_SUGGESTION_LENGTH
                        ):
                            # if line_index != matched_start_line:
                            # bp()
                            prev_lines = current_file_contents_lines[matched_start_line:line_index]
                            next_lines = current_file_contents_lines[line_index : predicted_end_line + 1]
                            rest_of_original_code_to_suggest = "\n".join(prev_lines + next_lines).strip("\n")
                            break

        except Exception as e:
            logger.warning(f"Could not parse code tree for file {self.current_file_path}:\n{str(e)}")
        return rest_of_original_code_to_suggest

    def fill_original_code_by_double_newline(self, partial_original_code: str) -> str:
        """
        Fills the original code by finding the next double newline
        """
        MIN_LINES_NEEDED_FOR_SUGGESTION = 10
        SUGGESTED_LINES = 50  # this is ideally the number of lines we want to suggest
        MAX_NAIVE_SUGGESTED_LINES = 150  # this is the maximum number of lines we can suggest naively
        current_file_contents = self.get_latest_contents(self.current_file_path)
        match_index = current_file_contents.index(partial_original_code)
        # this should just be the next tree-sitter node but we'll use double newline for now
        # Include the actual match and the following lines
        matched_content = current_file_contents[match_index + len(partial_original_code) :]
        lines = matched_content.split("\n")
        # we want the index of the furthest double newline
        next_double_newline_index = 0
        for i, line in enumerate(lines[:SUGGESTED_LINES]):
            if line.strip() == "" and i + 1 < len(lines) and len(lines[i + 1]) > 0:
                next_double_newline_index = i
        suggest_more_lines = False
        if next_double_newline_index <= MIN_LINES_NEEDED_FOR_SUGGESTION:
            next_double_newline_index = SUGGESTED_LINES
            suggest_more_lines = True
            # try again but this time with the maximum number of lines, break as soon as we can
            # it is better to suggest more and end on a double new line that to suggest less and end on a single new line
            for i, line in enumerate(lines[:MAX_NAIVE_SUGGESTED_LINES]):
                if line.strip() == "" and i + 1 < len(lines) and len(lines[i + 1]) > 0 and i > SUGGESTED_LINES:
                    next_double_newline_index = i
                    break
        suggested_lines = min(next_double_newline_index, SUGGESTED_LINES)
        if suggest_more_lines:
            suggested_lines = min(next_double_newline_index, MAX_NAIVE_SUGGESTED_LINES)
        # Take the match and up to suggested_lines after it
        result_lines = lines[: suggested_lines + 1]
        # iterate through result lines backwards and remove any lines that are pure whitespace
        for i in range(len(result_lines) - 1, -1, -1):
            if result_lines[i].strip() == "":
                result_lines.pop(i)
            else:
                break
        result = "\n".join(result_lines)
        return result

    def generate_suggested_response(self):
        """
        Attempt to generate a suggested response based on the current state
        """
        if not self.partial_code_change:
            self.handle_early_update_state_exit()
            return

        # TODO: handle missing original_code tags

        if self.partial_code_change.generating_original_code and self.partial_code_change.original_code:
            partial_original_code = self.partial_code_change.original_code
            # safe=True here because if current_file_contents is empty/nonexistent, and therefore partial_original_code is wrong, partial_original_code will be truncated, and LLM will continue to generate code as new_code
            current_file_contents = self.get_latest_contents(self.current_file_path, safe=True)
            MIN_LINES_NEEDED_FOR_SUGGESTION = 1
            # at least 2 lines of original code have to be generated before we can suggest the rest of the code, this is to prevent issues where ``` are the only thing generated
            if (
                len(self.partial_code_change.original_code.replace("`", "").splitlines())
                < MIN_LINES_NEEDED_FOR_SUGGESTION
            ):
                self.handle_early_update_state_exit()
                return

            # Try suggesting the rest of the code
            if current_file_contents:
                partial_original_code = self.validate_and_fix_partial_original_code(partial_original_code)
            else:
                # YOU ARE CREATING A NEW FILE HERE
                # if the file doesn't exist, truncate and return
                new_right_side = "\n```\n</original_code>\n<new_code>\n"
                self.suggested_response = self.current_response.rsplit(partial_original_code, 1)[0] + new_right_side
                return

            # We should truncate if the the partial original code is not in the current file contents
            if self.partial_code_change.generating_original_code and (replacement_original_code := get_replacement_original_code(current_file_contents, partial_original_code)):
                self.suggested_response = self.current_response.rsplit(partial_original_code, 1)[0] + replacement_original_code

            partial_original_code = self.validate_and_fix_partial_original_code(partial_original_code)

            # START OF ANCHOR TRUNCATION
            language = self.current_file_path.split(".")[-1]
            if boundary_truncated_code := postprocessed_truncate_on_boundary(
                current_file_contents, partial_original_code, language
            ):
                self.suggested_response = reset_rightmost_string(
                    self.current_response,
                    partial_original_code,
                    boundary_truncated_code,
                )
                # breakpoint()
                return
            # END OF ANCHOR_TRUNCATION

            # if there is a unique direct match, we can suggest the rest of the code
            original_code_to_suggest = ""
            if current_file_contents.count(partial_original_code) == 1:
                # try to generate the original code based on the entities in the file
                original_code_to_suggest = self.try_generate_original_code_with_entities(partial_original_code)
                if not original_code_to_suggest:
                    rest_of_original_code_to_suggest = self.fill_original_code_by_double_newline(partial_original_code)
                    if rest_of_original_code_to_suggest:
                        original_code_to_suggest = partial_original_code + rest_of_original_code_to_suggest

                if original_code_to_suggest:
                    if boundary_truncated_code := postprocessed_truncate_on_boundary(
                        current_file_contents, original_code_to_suggest, language
                    ):
                        original_code_to_suggest = boundary_truncated_code
                        # breakpoint()
                    original_code_to_suggest = reset_rightmost_string(
                        self.current_response,
                        partial_original_code,
                        original_code_to_suggest,
                    ).rstrip("\n")

                # final check to see if the suggested code is at the end of the file
                trailing_content = self.check_if_suggested_code_is_at_end_of_file(
                    current_file_contents, original_code_to_suggest
                )
                if trailing_content:
                    original_code_to_suggest += trailing_content
                if original_code_to_suggest:
                    self.suggested_response = original_code_to_suggest

        else:
            self.suggested_response = ""  # reset suggested response if there is nothing to suggest

    def consume_suggested_assistant_content(self) -> tuple[str, Generator]:
        """
        Consume the suggested original code and update the state, returns a stream
        """
        suggested_original_code = self.suggested_response
        self.current_response = self.suggested_response
        self.starting_response = self.current_response
        self._all_suggested_responses.append(self.suggested_response) # used for logging and debugging
        self.suggested_response = ""
        prefilled_assistant_message = Message(content=suggested_original_code, role="assistant")
        self.thread = collapse_final_assistant_messages(self.thread)
        # if there is already an assistant message in the last spot, replace it
        # otherwise, append the new assistant message
        if self.thread.messages[-1].role == "assistant":
            self.thread.messages[-1] = prefilled_assistant_message
        else:
            self.thread.messages.append(prefilled_assistant_message)
        prefix_string = prefilled_assistant_message.content
        new_stream = continuous_llm_calls.stream(
            username=self.username,
            thread=self.thread,
            content="",  # must be empty stream for prefilled assistant message
            stop_sequences=self.stop_sequences,
            model=self.model,
            max_tokens=4096,
            __version__=STREAM_VERSION,
        )

        return prefix_string, new_stream  # type: ignore

    def apply_code_change(self, code_change: PartialCodeChange):
        if code_change.completed and code_change not in self.applied_code_changes:
            # apply updates
            file_contents_to_update = self.get_latest_contents(code_change.file_path, safe=True)
            try:
                if code_change.original_code == "":
                    # append to end
                    file_contents_to_update += f"\n\n{code_change.new_code}"
                else:
                    file_contents_to_update = file_contents_to_update.replace(
                        code_change.original_code, code_change.new_code, 1
                    )
                self.modify_files_dict.update_new_contents(code_change.file_path, file_contents_to_update)
                self.applied_code_changes.add(code_change)
            except Exception as e:
                self.error_message += f"\n\nCould not apply code change to file {code_change.file_path}:\n{str(e)}"
                logger.error(self.error_message)

    def apply_last_code_change(self):
        partial_code_changes = self.get_partial_code_changes()
        if partial_code_changes:
            self.apply_code_change(partial_code_changes[-1])

    def update_state_with_new_response(self, new_response: str):
        """
        Call this everytime a new token is generated to update the state
        """
        self.current_response = new_response

        # parse partial suggested code_changes
        partial_code_changes = self.get_partial_code_changes()

        # in the case of no partial code changes, no further state updates are necessary
        if not partial_code_changes:
            self.handle_early_update_state_exit()
            return

        # we assume that only the last code change is partially generated and all prior ones are fully complete
        # now we see if we can provide a suggested response to the model based on the latest partial code change
        self.partial_code_change = partial_code_changes[-1]

        # be strict here to keep the model in check
        has_llm_generated_any_new_tokens = len(self.current_response) - len(self.starting_response) > 0
        if not has_llm_generated_any_new_tokens:
            self.handle_early_update_state_exit()
            return

        ### VALIDATION LOGIC
        # Detect large deletion diffs before applying changes
        # We always know there's at least one partial code change at this point.
        # We can check the last index where it's completed for potential large negative diffs.
        # Use last_validated_code_change instead of an attribute on PartialCodeChange because self.partial_code_change is continuously reinstantiated
        is_partial_code_change_validated = self.last_validated_code_change == self.partial_code_change
        if self.partial_code_change.new_code and self.partial_code_change.completed and not is_partial_code_change_validated:
            self.validate_deletion_diffs_and_suggest_response(file_path=self.partial_code_change.file_path)
            if self.suggested_response:
                return
            else:
                self.last_validated_code_change = self.partial_code_change
        ### END VALIDATION LOGIC
        # if the last code change is already completed, we don't need to do anything
        if self.partial_code_change.completed:
            self.handle_early_update_state_exit()
            return

        ### HANDLE STATE UPDATES
        # PARSE CURRENT FILE PATH
        modifying_previously_unchanged_file = False
        if self.partial_code_change.file_path != self.current_file_path:
            # we have moved onto a previously unchanged file, update the file path and clear the suggested response
            self.current_file_path = self.partial_code_change.file_path
            modifying_previously_unchanged_file = True
        # see if it is in our modify_files_dict
        if modifying_previously_unchanged_file:
            if self.current_file_path not in self.modify_files_dict:
                # update modify_files_dict
                try:
                    # try to save unnecessary calls to get_latest_contents
                    current_normed_path = os.path.normpath(self.current_file_path)
                    # use safe=True because it's a new file
                    current_file_contents = self.get_latest_contents(current_normed_path, safe=True)
                    self.modify_files_dict.add_or_update_file(
                        file_path=self.current_file_path,
                        original_contents=current_file_contents,
                        new_contents=current_file_contents,
                    )
                except FileNotFoundError:
                    # we should not see this error because we're using get_latest_contents with safe=True
                    logger.error(f"File {self.current_file_path} not found")
                except Exception as e:
                    self.error_message = f"Could not get contents of file {self.current_file_path}:\n{str(e)}"
                    logger.error(self.error_message)

        # do not continue if an error has occurred
        if self.error_message:
            self.handle_early_update_state_exit()
            return

        # TODO: MORE ROBUST UPDATE MODIFIED FILES DICT BASED ON PREVIOUS CODE CHANGES
        # currently assume each change can be applied right away
        for code_change in partial_code_changes[:-1]:
            self.apply_code_change(code_change)

        if self.error_message:
            self.handle_early_update_state_exit()
            return

        # Check here after this is done, we don't want to generally early exit here because new files might be created which always have empty original_code
        current_original_code_is_empty = (not bool(self.partial_code_change.original_code)) or (
            not len(self.partial_code_change.original_code) > 0
        )
        if current_original_code_is_empty:
            self.handle_early_update_state_exit()
            return

        self.generate_suggested_response()

    def validate_and_fix_partial_original_code(self, partial_original_code: str) -> str:
        # This is bad practicd, causes side effects
        current_file_contents = self.get_latest_contents(self.current_file_path)
        # double check the indentation on the provided original_code
        # 2 cases, 1 for single line one for multi line
        spaces_needed = 0
        if len(partial_original_code.splitlines()) == 1:
            if current_file_contents.count(partial_original_code) == 1:
                match_line_index = find_match_line_index(current_file_contents, partial_original_code)
                if match_line_index == -1:
                    return partial_original_code
                actual_line = current_file_contents.splitlines()[match_line_index]
                if partial_original_code != actual_line:
                    temp_partial_original_code = actual_line
                    self.current_response = reset_rightmost_string(
                        self.current_response,
                        partial_original_code,
                        temp_partial_original_code,
                    )
                    partial_original_code = temp_partial_original_code
        else:  # multiline match
            match_line_index, spaces_needed = find_match_with_leading_spaces(
                current_file_contents, partial_original_code
            )
        if spaces_needed > 0:
            # we need to add spaces to the original code
            temp_partial_original_code = "\n".join(
                " " * spaces_needed + line for line in partial_original_code.splitlines()
            )
            self.current_response = reset_rightmost_string(
                self.current_response, partial_original_code, temp_partial_original_code
            )
            partial_original_code = temp_partial_original_code
        self.current_response = strip_partial_closing_tags(self.current_response)
        partial_original_code = strip_partial_closing_tags(partial_original_code)
        return partial_original_code
    
    def stream(
        self,
        user_message: str,
        request_assistant_message_content: str, # should be set as constant
    ):
        # Main stream loop that handles switching streams when code suggestions are generated
        stream = continuous_llm_calls.stream(
            self.username,
            self.thread,
            content=user_message,
            assistant_message_content=request_assistant_message_content,
            stop_sequences=["</user_response>"],
            model=self.model,
            __version__=STREAM_VERSION,
        )

        result_string = ""
        prefix_string = request_assistant_message_content  # used during stream restarts while generating code suggestions
        while True:
            try:
                current_streamed_text = next(stream)
                if not current_streamed_text:
                    continue
                result_string = prefix_string + current_streamed_text
                yield result_string
                if self.suggested_response:
                    if detect_infinite_loop_in_code_in_suggested_responses(self._all_suggested_responses):
                        # OpenAI pre-fills can trigger this, will investigate later
                        raise Exception("Sweep has failed to make code changes. Please report this to support@sweep.dev.")
                    del stream
                    prefix_string, stream = self.consume_suggested_assistant_content()
                    logger.debug(f"Auto-suggested:\n{prefix_string}")
            except StopIteration:
                break


if __name__ == "__main__":
    cloned_repo = MockClonedRepo(
        _repo_dir="/tmp/sweep-internal",
        repo_full_name="sweepai/sweep-internal",
    )
    thread = Thread()
    modify_files_dict = FileModificationState()
    code_change_stream_state: CodeChangeStreamState = CodeChangeStreamState(
        username="test",
        model=LATEST_CLAUDE_MODEL,
        cloned_repo=cloned_repo,
        thread=thread,
        stop_sequences=[],
        modify_files_dict=modify_files_dict,
    )
    test_response = """
<code_change>
<file_path>
tests/utils/test_str_utils.py
</file_path>
<justification>
test
</justification>
<original_code>
```
if __name__ == "__main__":
    pytest.main([__file__])"""

    test_response_truncate = """
<code_change>
<file_path>
tests/utils/test_str_utils.py
</file_path>
<justification>
test
</justification>
<original_code>
```
import pytest
from typing import Iterable, Union
# rest of the code
"""
    # code_change_stream_state.update_state_with_new_response(test_response)
    code_change_stream_state.update_state_with_new_response(test_response_truncate)
    # bp()



import psutil
from fastapi import FastAPI
from fastapi.responses import JSONResponse

app = FastAPI()


@app.get("/health")
def health_check():
    cpu_usage = psutil.cpu_percent(interval=0.1)
    memory_info = psutil.virtual_memory()
    disk_usage = psutil.disk_usage("/")
    network_traffic = psutil.net_io_counters()

    status = {
        "status": "UP",
        "details": {
            "system_resources": {
                "cpu_usage": cpu_usage,
                "memory_usage": memory_info.used,
                "disk_usage": disk_usage.used,
                "network_traffic": {
                    "bytes_sent": network_traffic.bytes_sent,
                    "bytes_received": network_traffic.bytes_recv,
                },
            },
        },
    }

    return JSONResponse(status_code=200, content=status)
